# (c) Copyright Riverlane 2020-2025.
"""
This module implements a repetition code for quantum memory and stability experiments.
"""

from collections.abc import Callable

from deltakit_circuit import PauliX, PauliZ, Qubit
from deltakit_circuit._basic_maps import BASIS_TO_PAULI
from deltakit_circuit._basic_types import Coord2D, Coord2DDelta
from deltakit_circuit._qubit_identifiers import PauliGate
from deltakit_circuit.gates import PauliBasis

from deltakit_explorer.codes._css._css_code import CSSCode
from deltakit_explorer.codes._stabiliser import Stabiliser


class RepetitionCode(CSSCode):
    """
    Class representing an n-qubit repetition code
    for memory and stability experiments.
    Qubits are laid out in a line alternating between data and ancilla.
    Connectivity is only required between neighbouring pairs of qubits.
    See graph below for a distance 3 code.

    User can specify whether they want to implement a bit-flip code
    (with Z stabilisers) or a phase-flip code (with X stabilisers)

    This class also contains methods that help set up circuits for memory
    and stability experiments.

    .. code-block:: text

        0├    o --- Z/X --- o --- Z/X --- o
         |
         └----┴------┴------┴------┴------┴
              0      1      2      3      4

    Parameters
    ----------
    distance : int
        Minimum weight for a logical operator.
    stabiliser_type : PauliBasis
        Sets the stabiliser type for the repetition code e.g. PauliBasis.Z for a
        bit-flip code.
    use_ancilla_qubits : bool, optional
        Sets whether stabilisers have ancilla qubits for full circuit simulations.
        Setting this to False will compile the output circuit with MPP measurements
        for phenomenological noise. By default, True.
    use_looping_stabiliser : bool, optional
        Sets whether a looping stabiliser (between first and last data qubits)
        is defined. By default, False.
    default_schedule : bool, optional
        Whether to use the default schedule to use for syndrome extraction. If True,
        the syndrome extraction is performed on the qubit to the left of the ancilla
        first, and then on the qubit to the right. If False, it is the opposite way
        round. By default, True.
    odd_data_qubit_coords : bool, optional
        Whether the data qubits should have odd x-coordinates (True) or even (False).
        By default, False.
    """

    def __init__(
        self,
        distance: int,
        stabiliser_type: PauliBasis = PauliBasis.Z,
        use_ancilla_qubits: bool = True,
        use_looping_stabiliser: bool = False,
        default_schedule: bool = True,
        odd_data_qubit_coords: bool = False,
    ):
        self.distance = distance
        self._odd_data_qubit_coords = odd_data_qubit_coords
        self._check_stabiliser_type_is_valid(stabiliser_type)
        self._check_distance_at_least_two(self.distance)

        if default_schedule:
            self._schedule = (Coord2DDelta(-1, 0), Coord2DDelta(1, 0))
        else:
            self._schedule = (Coord2DDelta(1, 0), Coord2DDelta(-1, 0))

        self._data_qubits = self._calculate_data_qubits()
        self._stabiliser_ancilla_qubits = self._calculate_stabiliser_ancilla_qubits(
            use_looping_stabiliser
        )

        self._stabiliser_pauli = BASIS_TO_PAULI[stabiliser_type]
        stabilisers = self._calculate_stabilisers(use_ancilla_qubits)
        x_logical, z_logical = self._calculate_logical_operators()

        super().__init__(
            stabilisers=stabilisers,
            x_logical_operators=x_logical,
            z_logical_operators=z_logical,
            use_ancilla_qubits=use_ancilla_qubits,
        )

    @staticmethod
    def _check_stabiliser_type_is_valid(stabiliser_type: PauliBasis):
        """
        Checks to ensure a valid stabiliser type is inputted into the code.
        """
        if stabiliser_type not in (PauliBasis.X, PauliBasis.Z):
            msg = (
                f"{stabiliser_type} is unsupported, only PauliBasis.X and "
                "PauliBasis.Z are allowed."
            )
            raise ValueError(msg)

    @staticmethod
    def _check_distance_at_least_two(distance: int):
        """
        Check the repetition code has a distance that is at least two (for error
        detection).
        """
        if distance < 2:
            msg = "Code distance must be at least 2."
            raise ValueError(msg)

    def _calculate_data_qubits(self) -> set[Qubit]:
        """
        Calculate data qubits for the code.

        Returns
        -------
        Set[Qubit]
            Data qubits of the code.
        """
        delta = 1 if self._odd_data_qubit_coords else 0
        return {Qubit(Coord2D(2 * x + delta, 0)) for x in range(self.distance)}

    def _calculate_stabiliser_ancilla_qubits(
        self, use_looping_stabiliser
    ) -> set[Qubit]:
        """
        Calculate ancilla qubits used for constructing the code stabilisers.

        Returns
        -------
        Set[Qubit]
            Ancilla qubits used for constructing code stabilisers.
        """
        stabiliser_ancillas = {
            Qubit(Coord2D(2 * x + 1 + self._odd_data_qubit_coords, 0))
            for x in range(self.distance - 1)
        }
        if use_looping_stabiliser:
            x_coord = 0 if self._odd_data_qubit_coords else 2 * self.distance - 1
            stabiliser_ancillas.add(Qubit(Coord2D(x_coord, 0)))
        return stabiliser_ancillas

    def _calculate_stabilisers(self, use_ancilla_qubits) -> list[list[Stabiliser]]:
        """
        Get a full list of stabilisers to measure for the code.
        Stabilisers are pairs of ZZ (for the bit-flip code) or XX (phase-flip code)
        operators acting on data qubits adjacent to each ancilla.
        """

        stabilisers_first: list[Stabiliser] = []
        stabilisers_second: list[Stabiliser] = []

        for ancilla in self._stabiliser_ancilla_qubits:
            paulis = []
            for delta in self._schedule:
                (x, y) = ancilla.unique_identifier + delta
                qubit = Qubit(Coord2D(x % (2 * self.distance), y))
                paulis.append(self._stabiliser_pauli(qubit))
            stabiliser = Stabiliser(paulis=paulis, ancilla_qubit=ancilla)

            if use_ancilla_qubits or ((ancilla.unique_identifier[0] - 1) // 2) % 2 == 0:
                stabilisers_first.append(stabiliser)
            else:
                stabilisers_second.append(stabiliser)

        _sort_fun: Callable[[Stabiliser], int] = lambda stab: (  # noqa: E731
            stab.ancilla_qubit.unique_identifier.x + 1
        ) % (2 * self.distance)

        stabilisers_first.sort(key=_sort_fun)
        stabilisers_second.sort(key=_sort_fun)
        return [stabilisers_first + stabilisers_second]

    def _calculate_logical_operators(
        self,
    ) -> tuple[tuple[set[PauliGate]], tuple[set[PauliGate]]]:
        """Return logical operators for the repetition code."""
        if self._stabiliser_pauli == PauliZ:
            x_logical = {PauliX(qubit) for qubit in self._data_qubits}
            z_logical = {
                PauliZ(Qubit(Coord2D(1, 0)))
                if self._odd_data_qubit_coords
                else PauliZ(Qubit(Coord2D(0, 0)))
            }
        else:
            x_logical = {
                PauliX(Qubit(Coord2D(1, 0)))
                if self._odd_data_qubit_coords
                else PauliX(Qubit(Coord2D(0, 0)))
            }
            z_logical = {PauliZ(qubit) for qubit in self._data_qubits}

        return (x_logical,), (z_logical,)



from collections.abc import Sequence

import matplotlib.pyplot as plt
import numpy as np
import numpy.typing as npt
from deltakit_core.plotting.colours import RIVERLANE_PLOT_COLOURS
from matplotlib.axes import Axes
from matplotlib.figure import Figure

from deltakit_explorer.analysis import LogicalErrorProbabilityPerRoundResults


def _lep_interpolated(
    spam: float, leppr: float, rounds_interpolated: npt.NDArray[np.floating]
) -> npt.NDArray[np.floating]:
    """Computes logical error that would be obtained with the provided values.

    Uses the formula ``F = Fs * Fε**r`` where:

    - ``F`` is the expected fidelity of the computation,
    - ``Fs`` is the fidelity of SPAM-related operations,
    - ``Fε`` is the fidelity of one quantum error-correction round,
    - ``r`` is the number of quantum error-correction rounds performed.

    Each fidelity is obtained from the respective error probability with the formula
    ``f = (1 - 2 * e)`` where ``f`` is any of ``F``, ``Fs`` or ``Fε`` and ``e`` is any
    of logical error probability, logical error probability of a SPAM or logical error
    probability per round.
    """
    expected_fidelity = (1 - 2 * spam) * (1 - 2 * leppr) ** rounds_interpolated
    return (1 - expected_fidelity) / 2


def plot_logical_error_probability_per_round(
    leppr_data: LogicalErrorProbabilityPerRoundResults,
    num_rounds: npt.NDArray[np.int_] | Sequence[int],
    logical_error_probability: npt.NDArray[np.float64] | Sequence[float],
    logical_error_probability_stddev: npt.NDArray[np.float64] | Sequence[float] | None = None,
    *,
    num_sigmas: int = 3,
    fig: Figure | None = None,
    ax: Axes | None = None,
) -> tuple[Figure, Axes]:
    """Plot the logical error probability per round data and the fitted curve.

    Args:
        leppr_data (LogicalErrorProbabilityPerRoundResults):
            Data class containing logical error probability per round fit results.
        num_rounds (npt.NDArray[numpy.int_] | Sequence[int]):
            a sequence of integers representing the number of rounds used to get the
            corresponding results in ``num_failed_shots`` and ``num_shots``.
        logical_error_probability (npt.NDArray[numpy.float64] | Sequence[float]):
            a sequence of floats representing the logical error probabilities
            corresponding to the number of rounds in ``num_rounds``.
        logical_error_probability_stddev (npt.NDArray[numpy.float64] | Sequence[float] | None, optional):
            a sequence of floats representing the standard deviation of the logical
            error probabilities corresponding to the number of rounds in ``num_rounds``.
            If None, no error bars will be plotted. Default is None.
        num_sigmas (int): number of sigmas to consider when plotting error bars.
        fig (Figure | None, optional):
            a matplotlib Figure object to plot on. If None, a new figure will be created.
            Default is None.
        ax (Axes | None, optional):
            a matplotlib Axes object to plot on. If None, a new axes will be created.
            Default is None.

    Returns:
        tuple[Figure, Axes]: The matplotlib Figure and Axes objects containing the plot.

    Example:

        >>> from deltakit_explorer.analysis import (
        ...     calculate_lep_and_lep_stddev, compute_logical_error_per_round,
        ... )
        >>> num_failed_shots=[34, 151, 356]
        >>> num_shots=[500000] * 3
        >>> num_rounds=[2, 4, 6]
        >>> res = compute_logical_error_per_round(
        ...     num_failed_shots=num_failed_shots,
        ...     num_shots=num_shots,
        ...     num_rounds=num_rounds,
        ... )
        >>> lep, lep_stddev = calculate_lep_and_lep_stddev(
        ...     fails=num_failed_shots, shots=num_shots
        ... )
        >>> fig, ax = plot_logical_error_probability_per_round(
        ...     res,
        ...     num_rounds=num_rounds,
        ...     logical_error_probability=lep,
        ...     logical_error_probability_stddev=lep_stddev,
        ... )
    """
    if (fig is None) ^ (ax is None):
        msg = "The 'fig' and 'ax' parameters should either be both None or both set."
        raise ValueError(msg)

    if fig is None and ax is None:
        fig, ax = plt.subplots()

    assert ax is not None
    assert fig is not None

    lens = {len(num_rounds), len(logical_error_probability)}
    if logical_error_probability_stddev is not None:
        lens.add(len(logical_error_probability_stddev))
    if len(lens) > 1:
        msg = (
            "The lengths of 'num_rounds', 'logical_error_probability' and "
            "'logical_error_probability_stddev' must be the same. Got the following "
            f"lengths: {lens}."
        )
        raise ValueError(msg)

    isort = np.argsort(num_rounds)
    num_rounds = np.asarray(num_rounds)[isort]
    logical_error_probability = np.asarray(logical_error_probability)[isort]
    if logical_error_probability_stddev is not None:
        logical_error_probability_stddev = (
            num_sigmas * np.asarray(logical_error_probability_stddev)[isort]
        )

    # Plot the logical error probabilities
    ax.errorbar(
        num_rounds,
        logical_error_probability,
        yerr=logical_error_probability_stddev,
        fmt=".",
        color=RIVERLANE_PLOT_COLOURS[0],
        label=f"Logical error probabilities (±{num_sigmas}σ)"  # noqa: RUF001
    )
    # Plot the fitted logical error probability per round curve
    leppr, leppr_stddev = leppr_data.leppr, leppr_data.leppr_stddev
    spam, spam_stddev = leppr_data.spam_error, leppr_data.spam_error_stddev
    rounds_interpolated = np.linspace(num_rounds[0], num_rounds[-1], 200)
    lep_interpolated = _lep_interpolated(spam, leppr, rounds_interpolated)
    ax.plot(
        rounds_interpolated,
        lep_interpolated,
        label=f"Fit, ε={leppr:.4f} ± {num_sigmas * leppr_stddev:.4f} ({num_sigmas}σ)",  # noqa: RUF001
        color=RIVERLANE_PLOT_COLOURS[1]
    )

    # Add error band to logical error probability per round curve
    lep_interpolated_low = _lep_interpolated(
        spam - num_sigmas * spam_stddev,
        leppr - num_sigmas * leppr_stddev,
        rounds_interpolated
    )
    lep_interpolated_high = _lep_interpolated(
        spam + num_sigmas * spam_stddev,
        leppr + num_sigmas * leppr_stddev,
        rounds_interpolated
    )
    ax.fill_between(
        rounds_interpolated,
        np.clip(lep_interpolated_low, 0, 1),
        np.clip(lep_interpolated_high, 0, 1),
        color=RIVERLANE_PLOT_COLOURS[0],
        alpha=0.2
    )

    ax.set_title("Logical Error Probability Per Round Fit")
    ax.set_xlabel("Rounds")
    ax.set_ylabel("Logical Error Probability")
    ax.legend()

    return fig, ax



# (c) Copyright Riverlane 2020-2025.
"""Description of ``deltakit.explorer.types`` namespace here."""

from deltakit_explorer.types._data_string import DataString
from deltakit_explorer.types._experiment_types import (
                                                       QECExperiment,
                                                       QECExperimentDefinition,
)
from deltakit_explorer.types._types import (
                                                       BinaryDataType,
                                                       CircuitParameters,
                                                       Decoder,
                                                       DecodingResult,
                                                       DetectionEvents,
                                                       LeakageFlags,
                                                       MatrixSpecifications,
                                                       Measurements,
                                                       NoiseModel,
                                                       ObservableFlips,
                                                       PhysicalNoiseModel,
                                                       QubitCoordinateToDetectorMapping,
                                                       RAMData,
                                                       SI1000NoiseModel,
                                                       Sizes,
                                                       TypedData,
                                                       TypedDataFile,
                                                       TypedDataString,
)

# List only public members in `__all__`.
__all__ = [s for s in dir() if not s.startswith("_")]



# (c) Copyright Riverlane 2020-2025.
"""
This module contains standalone functions which help extract information from
stabilisers.
"""
from collections.abc import Iterable, Sequence

from deltakit_circuit import GateLayer, PauliX, PauliY, PauliZ, Qubit
from deltakit_circuit._basic_maps import PAULI_TO_CP
from deltakit_circuit._qubit_identifiers import PauliGate
from stim import PauliString, Tableau

from deltakit_explorer.codes._stabiliser import Stabiliser


def get_entangling_layer(
    stabilisers: Sequence[Stabiliser],
    layer_index: int,
) -> GateLayer:
    """
    Constructs one layer of entangling gates in the syndrome extraction
    circuit of the stabilisers provided.

    Parameters
    ----------
    stabilisers : Sequence[Stabiliser]
        Stabilisers taking place simultaneously in the syndrome extraction circuit.
    layer_index : int
        The entangling gate layer in the syndrome extraction circuit.

    Returns
    -------
    GateLayer
    """
    ctrl_gate_dict = PAULI_TO_CP

    gates = []
    for stab in stabilisers:
        if (pauli := stab.paulis[layer_index]) is not None:
            two_qubit_gate = ctrl_gate_dict[type(pauli)]
            gates.append(two_qubit_gate(stab.ancilla_qubit, pauli.qubit))

    return GateLayer(gates)


def _transform_stabiliser(
    stabiliser: Stabiliser,
    tableau_qubits: tuple[Qubit],
    transform_tableau: Tableau,
) -> Stabiliser:
    """
    Transforms the stabiliser provided with a unitary Clifford specified by a Stim
    tableau.

    Parameters
    ----------
    stabiliser : Stabiliser
        The stabiliser to transform.
    tableau_qubits : Tuple[Qubit]
        The qubits present in the tableau, indexed in the same way as they appear in
        it.
    transform_tableau : stim.Tableau
        The tableau representing the unitary Clifford to apply to the stabiliser.

    Returns
    -------
    Stabiliser
        The transformed stabiliser.
    """
    output_stabiliser = PauliString(len(tableau_qubits))
    for pauli in stabiliser.paulis:
        if pauli is None:
            continue
        qubit_ind = tableau_qubits.index(pauli.qubit)
        if isinstance(pauli, PauliX):
            output_pauli = transform_tableau.x_output(qubit_ind)
        elif isinstance(pauli, PauliY):
            output_pauli = transform_tableau.y_output(qubit_ind)
        else:
            output_pauli = transform_tableau.z_output(qubit_ind)
        output_stabiliser *= output_pauli
    return Stabiliser(
        paulis=[
            [
                None,
                PauliX(tableau_qubits[ind]),
                PauliY(tableau_qubits[ind]),
                PauliZ(tableau_qubits[ind]),
            ][pauli_num]
            for ind, pauli_num in enumerate(output_stabiliser)
            if pauli_num > 0
        ]
    )


def _get_data_qubits_from_stabilisers(
    stabilisers: Sequence[Sequence[Stabiliser]],
) -> tuple[Qubit, ...]:
    """
    Extracts and returns data qubits from the sequence of stabiliser
    sequences provided.
    """
    return tuple(
        sorted(
            {
                qubit
                for simultaneous_stabilisers in stabilisers
                for stabiliser in simultaneous_stabilisers
                for qubit in stabiliser.data_qubits
            },
            key=lambda qubit: qubit.unique_identifier,
        )
    )


def pauli_gates_to_stim_pauli_string(
    pauli_gates: Iterable[PauliGate | None],
    data_qubit_to_index_lookup: dict[Qubit, int],
) -> PauliString:
    r"""
    Given an iterable of `PauliGate`\ s, express it as a stim.PauliString. To do
    this, a `data_qubit_to_index_lookup` dictionary must also be provided, where
    each data qubit is mapped uniquely to an integer, for the purpose of turning
    `pauli_gates` into a `stim.PauliString`.

    Parameters
    ----------
    pauli_gates : Iterable[Optional[PauliGate]]
        Iterable of `PauliGate`, to be turned into a `stim.PauliString`. Allows
        some of these terms to be None (in which case they are just ignored) such
        that e.g. Stabiliser.paulis may be passed directly into this function.
    data_qubit_to_index_lookup : dict[Qubit, int]
        Dictionary containing a lookup of each data qubit to an integer. This is
        needed since qubit coordinates are usually `Coord2D`, so we must have an
        integer representation for converting to a `stim.PauliString`.

    Returns
    -------
    stim.PauliString
        `stim.PauliString` expression of `pauli_gates`.
    """

    if not all(
        pauli.qubit in data_qubit_to_index_lookup
        for pauli in pauli_gates
        if pauli is not None
    ):
        missing_qubits = {
            pauli.qubit for pauli in pauli_gates if pauli is not None
        }.difference(set(data_qubit_to_index_lookup.keys()))
        msg = (
            "data_qubit_to_index_lookup does not contain entries for"
            f" {missing_qubits} in pauli_gates"
        )
        raise ValueError(msg)
    paulistr = "*".join(
        pauli.stim_identifier + str(data_qubit_to_index_lookup[pauli.qubit])
        for pauli in pauli_gates
        if pauli is not None
    )

    # Adds trailing identity so that the total number of elements in the pauli strings
    # always equals to the number of qubits. This is needed because this trail is present
    # in a Tableau's pauli strings, and two pauli strings which have different trails
    # are not deemed equal by stim. For instance, PauliString("X") != PauliString("X_")
    # Note: The '+ 1' accounts for the +/- sign that gets added in front of the Pauli String
    # when it is initialised.
    return PauliString(
        str(PauliString(paulistr)).ljust(len(data_qubit_to_index_lookup) + 1, "_")
    )



# (c) Copyright Riverlane 2020-2025.
"""Module explores how the error suppression factor (lambda)
builds a connection between QuOps and code distance.

References
----------
- https://doi.org/10.48550/arXiv.2408.13687
"""

from collections.abc import Callable


def _equal_or_less_brute_force_search(
    func: Callable[[int], float],
    target: float,
    minimum: int,
    maximum: int,
    step: int,
) -> int | None:
    """Brute force search or the solution at on an interval.

    func(...) is (unfortunately) not always a monotonic function.
    The function searches for smallest D from [min, max],
    with func(D) <= target.

    Parameters
    ----------
        func (Callable): a descending function.
        target (float): a value to search for.
        minimum (int): minimum tested value of D.
        maximum (int): maximum tested value of D.
        step (int): next value to consider is D+step.

    Returns
    -------
        int | None:
            a value, which safisfies the search, or None.
    """
    while minimum < maximum:
        value = func(minimum)
        if value <= target:
            return minimum
        minimum += step
    return None


def _calculate_lep(lambda0: float, lambda_: float, distance: int, num_rounds: int) -> float:
    """Returns the probability of observing a logical error on a code of fixed
    distance after a number of rounds.

    It uses the formula in Section VI.B of Supplementary Information in
    https://doi.org/10.48550/arXiv.2408.13687 which is the sum of the probabilities
    of all ways of there being an odd number of errors in fixed number of rounds.
    """
    lep_per_round = lambda0 * lambda_ ** (-(distance + 1) / 2)
    # At `lep_per_round` << 1 this is be approximated as `lep_per_round * num_rounds`
    return 0.5 * (1 - (1 - 2 * lep_per_round) ** num_rounds)

def predict_quops_at_distance(lambda0: float, lambda_: float, distance: int) -> float:
    """Returns the number of QuOps, given distance.

    This uses the definition that the number of QuOps achievable is 1 / pL, where
    pL is the probability of a logical error occurring in distance-D, D-round block.

    Parameters
    ----------
    lambda0 (float):
        constant factor of lambda fit.
    lambda_ (float):
        Error suppression factor.
    distance (int):
        The distance at which to calculate the number of QuOps.
    """
    if distance % 2 == 0:
        msg = (
            "This method gives correct estimation only at odd distances. "
            f"Distance provided: {distance}"
        )
        raise ValueError(msg)
    return 1. / _calculate_lep(lambda0, lambda_, distance, distance)

def predict_distance_for_quops(
    lambda0: float,
    lambda_: float,
    num_quops: float,
    max_distance: int=999,
) -> int:
    """Returns the nearest odd distance that achieves the desired
    number of QuOps.

    Uses the definition that the number of QuOps achievable at a
    particular distance is 1 / pL, where pL is the probability of a logical error
    occurring during a distance-D, D-round memory experiment
    without state preparation or measurement error.

    Parameters
    ----------
    lambda0 (float):
        constant factor of lambda fit.
    lambda_ (float):
        Error suppression factor.
    num_quops (int):
        Number of desired QuOps, must be a positive integer greater than 2.
    max_distance (int):
        maximum distance to consider. Default is 999.

    Raises
    ------
    ValueError
        - if solution is not found;
        - num_quops < 2;
        - lambda_ <= 1.0
    """

    if num_quops < 2:
        msg = "Number of QuOps should be at least 2"
        raise ValueError(msg)

    if lambda_ <= 1.0:
        msg = "Lambda should be greater than 1 to ensure error suppression"
        raise ValueError(msg)

    required_lep = 1. / num_quops
    distance = _equal_or_less_brute_force_search(
        lambda x: _calculate_lep(lambda0, lambda_, x, x),
        required_lep,
        minimum=1,
        maximum=max_distance,
        step=2,  # seek for odd solutions
    )
    if distance is None:
        text = (
            f"Could not find a solution between [1, {max_distance}] "
            "for LEP(distance) < 1 / QuOps. Increase max_distance."
        )
        raise ValueError(text)
    return distance



# (c) Copyright Riverlane 2020-2025.
"""`visualisation` module aggregates data plotting methods.
"""

from __future__ import annotations

from collections.abc import Collection, Iterable, Sequence
from itertools import chain

import matplotlib.pyplot as plt
import numpy as np
import numpy.typing as npt
import seaborn as sns
from matplotlib.ticker import FuncFormatter

from deltakit_explorer.types._types import QubitCoordinateToDetectorMapping


def correlation_matrix(
    matrix: npt.NDArray,
    qubit_to_detector_mapping: QubitCoordinateToDetectorMapping,
    labels: Sequence[str] = (),
):
    """Plot a given correlation matrix as a heatmap.

    Args:
        matrix (npt.NDArray): correlation matrix.
        qubit_to_detector_mapping (QubitCoordinateToDetectorMapping):
            {qubit_coordinate_tuple_1: [det_1, det_2, det_2, ...], }
        labels (Sequence[str]): labels to the qubits.

    Returns:
        matplotlib.plt:
            The plt object containing the drawn heatmap.

    Examples:

        Collect the data for a correlation matrix plot::

            matrix, mapping = client.get_correlation_matrix(
                detectors, stim_circuit,
                use_default_noise_model_edges=True,
            )
            plt = correlation_matrix(matrix, mapping)
            plt.show()

    """
    # create a list of indices of the minor ticks for which to label with
    # the qubit labels such that the labels are in the middle of the major
    # ticks. Sort the labels as they are not guaranteed to be in order.
    minor_ticks_in_major = len(
        next(iter(qubit_to_detector_mapping.detector_map.values())))
    num_major_ticks = len(qubit_to_detector_mapping.detector_map.keys())
    num_ticks = minor_ticks_in_major * num_major_ticks
    num_minor_ticks = num_ticks - num_major_ticks
    ticks_per_major = num_minor_ticks // num_major_ticks
    mid_im = ticks_per_major // 2
    label_indices = [mid_im + (ticks_per_major * i) for i in range(num_major_ticks)]
    sorted_labels = (
        sorted(qubit_to_detector_mapping.detector_map.keys())
        if len(labels) == 0 else labels
    )

    def format_func(_, tick_number):
        if tick_number in label_indices:
            return sorted_labels[label_indices.index(tick_number)]
        return None

    col1 = sns.cubehelix_palette(start=2.43, rot=-0.1, light=1.0, as_cmap=True)
    axes = sns.heatmap(matrix, cmap=col1)
    axes.invert_yaxis()
    axes.set(xlabel="Qubit coordinate", ylabel="Qubit coordinate")
    major_ticks = np.arange(0, len(matrix[0]), minor_ticks_in_major)
    minor_ticks = np.arange(0, len(matrix[0]), 1)
    axes.set_xticks(major_ticks)
    axes.set_xticks(minor_ticks, minor=True)
    axes.set_yticks(major_ticks)
    axes.set_yticks(minor_ticks, minor=True)
    axes.xaxis.set_minor_formatter(FuncFormatter(format_func))
    axes.yaxis.set_minor_formatter(FuncFormatter(format_func))
    axes.tick_params(axis="x", rotation=0)
    axes.grid(which="major", color="#333333", linestyle="-", alpha=0.6)
    axes.grid(which="minor", color="#AAAAAA", linestyle="--", alpha=0.2)
    return plt

def _rotate_defect_rate_points(
    detector_coords: dict,
    defect_rates: dict,
    transform_mat=np.array([[1, 1], [-1, 1]]) * (0.5**0.5),
) -> dict:
    # transform det coordinates to get upright plot
    for d_id, coord in detector_coords.items():
        detector_coords[d_id] = [*list(transform_mat @ coord[:-1]), coord[-1]]

    # transform defect rate coords
    x_offset = 0
    y_offset = 0
    rotated_coord_defect_rates = {}
    for coord, rates in defect_rates.items():
        new_coord = tuple((transform_mat @ coord).round())
        x_offset = min(x_offset, new_coord[0])
        y_offset = min(y_offset, new_coord[1])
        rotated_coord_defect_rates[new_coord] = np.mean(rates[1:-1])

    offset_defect_rates = {}
    for coord, defect_rate in rotated_coord_defect_rates.items():
        offset_defect_rates[
            (coord[0] - x_offset, coord[1] - y_offset)
        ] = defect_rate

    return offset_defect_rates

def defect_diagram(all_detector_coords: dict, all_defect_rates: dict):
    """Plots defect rates patch diagram given detector coordinates and
    their error rates.

    Args:
        all_detector_coords (Dict): Mapping from coordinates to detector numbers.
        all_defect_rates (Dict): Defect rates of detectors.

    Returns:
        matplotlib.pyplot: matplotlib module
    """
    # rotate coords
    defect_rates = _rotate_defect_rate_points(
        all_detector_coords, all_defect_rates
    )
    all_dr_means = defect_rates.values()
    cmap_min, cmap_max = min(all_dr_means), max(all_dr_means)

    all_coords = list(chain.from_iterable(defect_rates.keys()))

    # plot numbers in matrix
    matrix_width = max(all_coords)
    tpl: tuple[int, ...] = (int(matrix_width) + 1, int(matrix_width) + 1)
    defect_rate_mat = np.zeros(tpl, dtype=np.float64)
    for coord, val in defect_rates.items():
        defect_rate_mat[round(coord[0]), round(coord[1])] = val
    defect_rate_mat = np.array(defect_rate_mat).T

    # remove all-0 columns and rows
    defect_rate_mat = defect_rate_mat[~np.all(defect_rate_mat == 0, axis=1)]
    zero_column_idx = np.argwhere(np.all(defect_rate_mat[..., :] == 0, axis=0))
    defect_rate_mat = np.delete(defect_rate_mat, zero_column_idx, axis=1)

    # get coords to plot semicircles:
    # to plot these we need to find the weight-2 stabiliser values in the heatmap.
    # these will exist in the first and last rows and columns.
    # find their values + coordinates, and then shift them to accommodate
    # for deleting the first and last rows and columns, as we instead want to plot
    # semicircles in place of the squares.
    num_rows, num_cols = defect_rate_mat.shape
    last_col, first_col = defect_rate_mat[:, 0], defect_rate_mat[:, num_cols - 1]
    last_row, first_row = defect_rate_mat[0, :], defect_rate_mat[num_rows - 1, :]

    # create coordinates where the circles will be drawn, and keep a record
    # of the value so that the circles can be plotted with the correct colour
    top_sc_indices = [
        (num_rows - 1 - 1 - 0.5, x - 1, first_row[x])
        for x in np.where(first_row != 0)[0]
    ]
    bottom_sc_indices = [
        (-0.5, x - 1, last_row[x]) for x in np.where(last_row != 0)[0]
    ]
    left_sc_indices = [
        (num_cols - x - 1 - 1, -0.5, first_col[x])
        for x in np.where(first_col != 0)[0]
    ]
    right_sc_indices = [
        (num_cols - x - 1 - 1, num_cols - 1 - 1 - 0.5, last_col[x])
        for x in np.where(last_col != 0)[0]
    ]

    # delete rows/columns containing weight-2 stabilisers
    # to make room for the semicircles
    defect_rate_mat = np.delete(defect_rate_mat, (0, num_cols - 1), axis=1)
    defect_rate_mat = np.delete(defect_rate_mat, (0, num_rows - 1), axis=0)

    # begin plot
    fig, axes = plt.subplots(figsize=(5, 5))
    cmap = plt.get_cmap("viridis")

    # plot heatmap+colorbar, hide axis
    image = axes.imshow(defect_rate_mat, cmap=cmap, vmin=0.1, vmax=0.25, zorder=1)
    axes.set_yticks([])
    axes.set_xticks([])
    image.axes.invert_yaxis()
    cbar = axes.figure.colorbar(
        image, ax=axes, shrink=1, orientation="vertical",
        pad=0.2, label="Defect rate"
    )
    cbar.set_ticks([0.1, 0.25])

    # plot semicircles
    for circles in (
        top_sc_indices,
        bottom_sc_indices,
        left_sc_indices,
        right_sc_indices,
    ):
        for circle_coord in circles:
            rgba = cmap((circle_coord[2] - cmap_min) / cmap_max)
            circle = plt.Circle(
                (circle_coord[1], circle_coord[0]),
                0.5,
                color=rgba,
                clip_on=False,
                zorder=0,
            )
            axes.add_patch(circle)

    # plot
    fig.tight_layout()
    return plt


def defect_rates(
    defect_rates_series: Iterable[dict[tuple[float, ...], list[float]]],
    w2_det_coords: Collection[tuple[float, ...]],
):
    """
    Convenience function to plot the average defect rate plot fig 2b
    in Google paper https://www.nature.com/articles/s41586-022-05434-1.

    Args:
        defect_rates_series (Iterable[Dict[Tuple[float, ...], List[float]]]):
            List of defect rates dictionaries.
            E.g, this can be for the X and Z experiments
            for the Google data set.
        w2_det_coords (Container[Tuple[float, ...]]):
            Coordinates for the weight 2 detectors, so
            that these may be plotted with a separate colour
            and have their average separate from the higher-weight
            stabilisers.

    Returns:
        matplotlib.pyplot: pyplot module

    Examples:
        Plotting the Google example::

            z_data_folder = data_folder / "surface_code_bZ_d3_r07_center_3_5/"
            x_data_folder = data_folder / "surface_code_bX_d3_r07_center_3_5/"
            z_and_x_experiment_folders = [z_data_folder, x_data_folder]

            experiments = [
                QECExperiment.from_circuit_and_measurements(
                    folder / "circuit_noisy.stim",
                    folder / "measurements.b8", DataFormat.B8,
                    folder / "sweep.b8", DataFormat.B8,
                )
                for folder
                in z_and_x_experiment_folders
            ]
            all_rates = []
            for experiment in experiments:
                _, rates = client.get_experiment_detectors_and_defect_rates(
                    experiment
                )
                all_rates.append(rates)
            defect_rates(
                all_rates,
                w2_det_coords=set({(5., 6.), (1., 4.), (4., 3.), (2., 7.)})
            )
    """
    # ensure these are floats!
    w2_det_coords = {tuple(map(float, item)) for item in w2_det_coords}
    w2_avg = []
    w4_avg = []
    for all_qubit_defect_rates in defect_rates_series:
        for coord, defect_rate in all_qubit_defect_rates.items():
            if coord in w2_det_coords:
                w2_avg.append(defect_rate)
                plt.plot(
                    range(1, len(defect_rate) + 1),
                    defect_rate, color="#ff7500", alpha=0.3
                )
            else:
                w4_avg.append(defect_rate)
                plt.plot(
                    range(1, len(defect_rate) + 1),
                    defect_rate, color="#006f62", alpha=0.3
                )
    w2_detectors = np.mean(w2_avg or [[]], axis=0)
    w4_detectors = np.mean(w4_avg or [[]], axis=0)
    plt.plot(
        range(1, len(w4_detectors) + 1),
        w4_detectors,
        color="#006f62", label="Weight-4")
    plt.plot(
        range(1, len(w2_detectors) + 1),
        w2_detectors,
        color="#ff7500", label="Weight-2")
    plt.xlabel("Round")
    plt.xticks(range(1, len(w4_detectors) + 1))
    plt.ylabel("Defect rate")
    plt.legend(loc="lower right", frameon=False)
    return plt



# (c) Copyright Riverlane 2020-2025.
"""Enums used across the client library."""
from __future__ import annotations

from enum import Enum


class APIEndpoints(Enum):
    """Names of API endpoints"""

    GENERATE_CIRCUIT = "generate_circuit"
    ADD_NOISE = "generate_noisy_stim_circuit"
    ADD_SI1000_NOISE = "generate_si1000_noisy_stim_circuit"
    DECODE = "decode"
    DECODE_LEAKAGE = "decode_leakage"
    DEFECT_RATES = "defect_rates"
    CORRELATION_MATRIX = "correlation_matrix"
    TRIM_CIRCUIT_AND_DETECTORS = "trim_circuit_and_detectors"
    SIMULATE_CIRCUIT = "simulation"


class DataFormat(Enum):
    """Represent common data representation formats.

    Supported values are ``F01``, ``B8``, ``CSV``, and ``TEXT``.
    """

    F01 = "01"
    B8 = "b8"
    CSV = "csv"
    TEXT = "text"


class DecoderType(Enum):
    """Decoder types, supported by a client. Supported values are:

    - ``MWPM``: Minimum Weight Perfect Matching [1]_
    - ``CC``: Collision Clustering [2]_
    - ``BELIEF_MATCHING``: Belief Matching [3]_
    - ``BP-OSD``: Belief Propagation - Ordered Statistics Decoding [4]_
    - ``AC``: Ambiguity Clustering [5]_
    - ``LCD``: Local Clustering Decoder [6]_

    References
    ----------
    .. [1] https://arxiv.org/abs/2303.15933
    .. [2] https://arxiv.org/abs/2309.05558
    .. [3] https://journals.aps.org/prx/abstract/10.1103/PhysRevX.13.031007
    .. [4] https://quantum-journal.org/papers/q-2021-11-22-585/
    .. [5] https://arxiv.org/abs/2406.14527
    .. [6] https://arxiv.org/abs/2411.10343

    """

    MWPM = "MWPM"
    CC = "CC"
    BELIEF_MATCHING = "BELIEF_MATCHING"
    BP_OSD = "BP_OSD"
    AC = "AC"
    LCD = "LCD"


class QECExperimentType(Enum):
    """Type of QEC experiment.

    Supported values are ``QUANTUM_MEMORY`` and ``STABILITY``.

    """

    QUANTUM_MEMORY = "QUANTUM_MEMORY"
    STABILITY = "STABILITY"


class QECECodeType(Enum):
    """Type of QEC code.

    Supported values are ``ROTATED_PLANAR``, ``UNROTATED_PLANAR``,
    ``REPETITION``, ``BIVARIATE_BICYCLE``, and ``UNROTATED_TORIC``.

    """

    ROTATED_PLANAR = "ROTATED_PLANAR"
    UNROTATED_PLANAR = "UNROTATED_PLANAR"
    REPETITION = "REPETITION"
    BIVARIATE_BICYCLE = "BIVARIATE_BICYCLE"
    UNROTATED_TORIC = "UNROTATED_TORIC"



# (c) Copyright Riverlane 2020-2025.
# This module is not currently public; this __init__.py file is a historical artefact
# and can be removed, adjusting imports within other `deltakit_explorer` modules
# accordingly.
from deltakit_explorer.qpu._noise._noise_parameters import NoiseParameters
from deltakit_explorer.qpu._noise._phenomenological_noise import PhenomenologicalNoise
from deltakit_explorer.qpu._noise._sd6_noise_model import SD6Noise
from deltakit_explorer.qpu._noise._si1000_noise import SI1000Noise
from deltakit_explorer.qpu._noise._toy_noise import ToyNoise

__all__ = [
    "NoiseParameters",
    "PhenomenologicalNoise",
    "SD6Noise",
    "SI1000Noise",
    "ToyNoise",
]



from __future__ import annotations

import warnings
from collections.abc import Callable, Sequence
from dataclasses import dataclass
from typing import Literal

import numpy as np
import numpy.typing as npt
import scipy.optimize


@dataclass(frozen=True)
class LambdaResults:
    """Named-tuple-like class containing computation results from
    :func:`calculate_lambda_and_lambda_stddev`.

    Attributes:
        lambda_ (float): computed error suppression factor.
        lambda_stddev (float): lambda standard deviation.
        lambda0 (float): computed error suppression multiplicative offset (value of Λ_0
            in the expression ``Ɛ_d = 1 / [ Λ_0 * Λ**((d+1)/2) ]``).
        lambda0_stddev (float): Λ_0 standard deviation.
    """

    lambda_: float
    lambda_stddev: float
    lambda0: float
    lambda0_stddev: float


_LambdaFittingCallable = Callable[
    [
        npt.NDArray[np.int_] | Sequence[int],
        npt.NDArray[np.float64] | Sequence[float],
        npt.NDArray[np.float64] | Sequence[float],
    ],
    LambdaResults,
]


def _lambda_fit_with_d(
    distances: npt.NDArray[np.int_] | Sequence[int],
    lep_per_round: npt.NDArray[np.float64] | Sequence[float],
    lep_stddev_per_round: npt.NDArray[np.float64] | Sequence[float],
) -> LambdaResults:
    """Compute Λ, Λ_0 and their associated standard deviations by fitting the logarithm
    of ``lep_per_round`` with ``distance``.
    """
    # Prepare data for the fit.
    lep_per_round = np.asarray(lep_per_round, dtype=np.float64)
    logleppr = np.log(lep_per_round)
    logleppr_stddev = lep_stddev_per_round / lep_per_round
    # Fitting with numpy.polyfit to be able to provide standard deviations and recover a
    # covariance matrix as numpy.polynomial.Polyfit is not able to do that yet.
    (slope, offset), cov = np.polyfit(
        distances, logleppr, 1, w=1 / logleppr_stddev, full=False, cov="unscaled"
    )
    slope_stddev, offset_stddev = np.sqrt(np.diagonal(cov))
    # Recovering the numbers of interest. Maths representing what has been performed:
    # We start from Ɛ_d = 1 / [ Λ_0 * Λ**((d+1)/2) ]
    # Applying ln:  ln(Ɛ_d) = - ln(Λ_0) - (d+1)/2 * ln(Λ)
    #                       = - ln(Λ_0) - ln(Λ)/2 - d * ln(Λ)/2
    # The linear fit performed above gave us slope  = -ln(Λ)/2
    #                                        offset = -ln(Λ_0) - ln(Λ)/2
    lambda_value = float(np.exp(-2 * slope))
    lambda_value_stddev = float(lambda_value * 2 * slope_stddev)
    lambda0 = float(np.exp(-offset - np.log(lambda_value) / 2))
    # Λ_0 = exp(-offset - ln(Λ)/2)
    # Error analysis (to compute the standard deviation of Λ_0) done with the formulas
    # in https://en.wikipedia.org/wiki/Propagation_of_uncertainty#Example_formulae:
    # σ(ln(Λ)/2) = σ(Λ) / (2 * Λ)
    # σ(offset) is obtained from the covariance matrix
    # σ(-offset - ln(Λ)/2) = √(σ(offset)² + σ(ln(Λ) / 2)²
    #                          - 2 * covariance(offset, ln(Λ) / 2))
    #                      = √(σ(offset)² + σ(Λ)² / (4 * Λ²)
    #                          - 2 * covariance(offset, ln(Λ) / 2))
    # σ(exp(-offset - ln(Λ)/2)) = exp(-offset - ln(Λ)/2) * σ(-offset - ln(Λ)/2)
    #                           = Λ_0 * √(σ(offset)² + σ(Λ)² / (4 * Λ²)
    #                                     - 2 * covariance(offset, ln(Λ) / 2))
    lambda0_stddev = float(
        lambda0
        * np.sqrt(
            offset_stddev**2
            + lambda_value_stddev**2 / (4 * lambda_value**2)
            - 2 * cov[0, 1]
        )
    )
    return LambdaResults(lambda_value, lambda_value_stddev, lambda0, lambda0_stddev)


def _lambda_fit_with_d_plus_1_over_2(
    distances: npt.NDArray[np.int_] | Sequence[int],
    lep_per_round: npt.NDArray[np.float64] | Sequence[float],
    lep_stddev_per_round: npt.NDArray[np.float64] | Sequence[float],
) -> LambdaResults:
    """Compute Λ, Λ_0 and their associated standard deviations by fitting the logarithm
    of ``lep_per_round`` with ``(distance + 1) / 2``.
    """
    # Prepare data for the fit.
    distances = np.asarray(distances, dtype=np.int_)
    lep_per_round = np.asarray(lep_per_round, dtype=np.float64)
    logleppr = np.log(lep_per_round)
    logleppr_stddev = lep_stddev_per_round / lep_per_round
    # Fitting with numpy.polyfit to be able to provide standard deviations and recover a
    # covariance matrix as numpy.polynomial.Polyfit is not able to do that yet.
    (slope, offset), cov = np.polyfit(
        (distances + 1) / 2,
        logleppr,
        1,
        w=1 / logleppr_stddev,
        full=False,
        cov="unscaled",
    )
    slope_stddev, offset_stddev = np.sqrt(np.diagonal(cov))
    # Recovering the numbers of interest. Maths representing what has been performed:
    # We start from Ɛ_d = 1 / [ Λ_0 * Λ**((d+1)/2) ]
    # Applying ln:  ln(Ɛ_d) = - ln(Λ_0) - (d+1)/2 * ln(Λ)
    # The linear fit performed above gave us slope  = -ln(Λ)
    #                                        offset = -ln(Λ_0)
    lambda_value = float(np.exp(-slope))
    lambda_value_stddev = float(lambda_value * slope_stddev)
    lambda0 = float(np.exp(-offset))
    lambda0_stddev = float(lambda0 * offset_stddev)
    return LambdaResults(lambda_value, lambda_value_stddev, lambda0, lambda0_stddev)


def _lambda_fit_with_direct(
    distances: npt.NDArray[np.int_] | Sequence[int],
    lep_per_round: npt.NDArray[np.float64] | Sequence[float],
    lep_stddev_per_round: npt.NDArray[np.float64] | Sequence[float],
) -> LambdaResults:
    """Compute Λ, Λ_0 and their associated standard deviations by fitting
    ``lep_per_round`` to ``1 / Λ_0 * Λ**(-(distance + 1) / 2)`` directly.

    This method does not rely on least-square polynomial fitting but rather on a more
    generic method. As such, it requires more time to converge.
    """
    # Prepare data for the fit.
    distances = np.asarray(distances, dtype=np.int_)
    # Here we are not fitting a polynomial anymore but directly the formula:
    #   Ɛ_d = 1 / [ Λ_0 * Λ**((d+1)/2) ]
    # with ``x`` that is ``(d+1)/2``.
    (lamb0, lamb), cov = scipy.optimize.curve_fit(
        lambda x, lamb0, lamb: 1 / lamb0 * lamb ** (-x),
        (distances + 1) / 2,
        lep_per_round,
        sigma=lep_stddev_per_round,
        absolute_sigma=True,
        jac=lambda x, lamb0, lamb: np.transpose(
            [
                -1 / lamb0**2 * lamb ** (-x),
                -1 / lamb0 * x * lamb ** (-x - 1),
            ]
        ),
        # Both parameters below are needed for crazy values of lambda and lambda0 to
        # make sure the method converges to the correct value.
        bounds=(0, np.inf),
        maxfev=10000,
    )
    lamb0_stddev, lamb_stddev = np.sqrt(np.diagonal(cov))
    return LambdaResults(
        float(lamb), float(lamb_stddev), float(lamb0), float(lamb0_stddev)
    )


_LAMBDA_FITTING_METHODS: dict[
    Literal["d", "(d+1)/2", "direct"], _LambdaFittingCallable
] = {
    "d": _lambda_fit_with_d,
    "(d+1)/2": _lambda_fit_with_d_plus_1_over_2,
    "direct": _lambda_fit_with_direct,
}


def calculate_lambda_and_lambda_stddev(
    distances: npt.NDArray[np.int_] | Sequence[int],
    lep_per_round: npt.NDArray[np.float64] | Sequence[float],
    lep_stddev_per_round: npt.NDArray[np.float64] | Sequence[float],
    method: Literal["d", "(d+1)/2", "direct"] = "(d+1)/2",
) -> LambdaResults:
    """Calculate the error suppression factor (Λ) and its standard deviation.

    Requires the logical error probability (LEP) per round (which may be approximated
    as LEP / num_rounds for small LEP or computed with
    :func:`compute_logical_error_per_round` for a more precise approximation), and its
    standard deviation (also returned by :func:`compute_logical_error_per_round`).

    By providing the logical error probability for increasing code distances,
    one can obtain an estimate for how error suppression scales with distances.
    Note that lambda is a "rule of thumb". This approximation is unreliable near
    threshold and for low code distances. If such a regime is detected, a warning will
    be emitted by this function.

    Args:
        distances (npt.NDArray[numpy.int\\_] | Sequence[int]): Distances at which
            ``lep_per_round`` and ``lep_stddev_per_round`` are provided. Should only
            contain odd distances. Estimations of Λ may be unreliable when data from
            distance 3 is used and the value of Λ is low (see Fig. S15 of Supplementary
            information of "Quantum error correction below the surface code threshold"
            at https://www.nature.com/articles/s41586-024-08449-y#Sec8). If such a
            situation is encountered, a warning will be emitted.
        lep_per_round (npt.NDArray[numpy.float64] | Sequence[float]):
            logical error probabilities per round computed for each code distance in
            ``distances``. Should be the same size as ``distances``.
        lep_stddev_per_round (npt.NDArray[numpy.float64] | Sequence[float]):
            standard deviation of the logical error probabilities per round computed for
            each code distance in ``distances``. Should be the same size as
            ``distances``.
        method (Literal["d", "(d+1)/2", "direct"]): mathematical method used to fit the
            data. Defaults to "(d+1)/2". All 3 methods show remarkable numerical
            agreement, but "direct" is slower than both "d" and "(d+1)/2", so these last
            2 should be preferred in general.

    Returns:
        LambdaResults: detailed results of the computation.

    Note:
        For values of Λ very close to 1 (``abs(Λ - 1) < 1e-7``) and
        ``method == "direct"``, this function might emit a
        ``scipy.optimize._optimize.OptimizeWarning`` with the message ``"Covariance of
        the parameters could not be estimated"``.

        Realistically, that condition is not expected to occur in practice due to
        sampling noise and sampling overhead, but it might be checked by synthetic
        data (e.g., in unit-tests).

    Examples:
        Fitting the Λ value given information for 5, 7, and 9 round of a QEC
        experiment::

            res = calculate_lambda_and_lambda_stddev(
                distances=[5, 7, 9],
                lep_per_round=[1.992e-04, 4.314e-05, 7.556e-06],
                lep_stddev_per_round=[1.2e-05, 9.3e-06, 3.9e-06],
            )
            lambda_, lambda_stddev = res.lambda_, res.lambda_stddev

    """
    # Make sure that the inputs are numpy arrays sorted by distance
    isort = np.argsort(distances)
    distances = np.asarray(distances)[isort]
    lep_per_round = np.asarray(lep_per_round)[isort]
    lep_stddev_per_round = np.asarray(lep_stddev_per_round)[isort]

    # Check that we do not have duplicate data for the same distance as that will
    # confuse the numerical methods used in this function.
    unique_counts = np.unique_counts(distances)
    non_unique_entries_mask = unique_counts.counts > 1
    if np.any(non_unique_entries_mask):
        non_unique_values = unique_counts.values[non_unique_entries_mask].tolist()
        msg = (
            "Multiple entries were provided for the following distances: "
            f"{non_unique_values}. This is not supported."
        )
        raise ValueError(msg)

    # Make sure that there are no even distances.
    if np.any(distances % 2 == 0):
        msg = (
            "Found at least one even distance in the provided distances "
            f"({distances.tolist()}). This is not supported."
        )
        raise ValueError(msg)

    if method not in _LAMBDA_FITTING_METHODS:
        warnings.warn(
            "Got a fitting method that is not supported by this function "
            f"('{method}'). Valid methods are {list(_LAMBDA_FITTING_METHODS)}."
        )
    lambda_fit_func: _LambdaFittingCallable = _LAMBDA_FITTING_METHODS.get(
        method, _lambda_fit_with_d
    )
    res = lambda_fit_func(distances, lep_per_round, lep_stddev_per_round)
    if res.lambda_ < 1.5 and min(distances) < 5:
        warnings.warn(
            "Lambda estimation is unreliable at low code distances and low values of "
            "lambda. Please use distance 5 as a minimum.",
        )
    return res



# (c) Copyright Riverlane 2020-2025.
from deltakit_explorer.codes._bivariate_bicycle_code import (
    BivariateBicycleCode,
    Monomial,
    Polynomial,
)
from deltakit_explorer.codes._css._css_code import CSSCode
from deltakit_explorer.codes._css._css_code_experiment_circuit import (
    css_code_memory_circuit,
    css_code_stability_circuit,
)
from deltakit_explorer.codes._css._css_stage import CSSStage
from deltakit_explorer.codes._css._experiment_circuit import experiment_circuit
from deltakit_explorer.codes._css._stabiliser_code import StabiliserCode
from deltakit_explorer.codes._planar_code._planar_code import ScheduleType
from deltakit_explorer.codes._planar_code._rotated_planar_code import RotatedPlanarCode
from deltakit_explorer.codes._planar_code._unrotated_planar_code import (
    UnrotatedPlanarCode,
)
from deltakit_explorer.codes._planar_code._unrotated_toric_code import (
    UnrotatedToricCode,
)
from deltakit_explorer.codes._repetition_code import RepetitionCode
from deltakit_explorer.codes._schedules._schedule_order import ScheduleOrder

# List only public members in `__all__`.
__all__ = [s for s in dir() if not s.startswith("_")]



# (c) Copyright Riverlane 2020-2025.
"""
This module includes noise profiles to add leakage to noise models.
Defines noise profiles for leakage and relaxation contexts parameterised
on leakage (pL) and relaxation (pR) probabilities parameters.
"""

from deltakit_circuit import NoiseProfile
from deltakit_circuit.gates import (
    TWO_QUBIT_GATES,
    Gate,
    OneQubitCliffordGate,
    OneQubitMeasurementGate,
    OneQubitResetGate,
)
from deltakit_circuit.noise_channels import Depolarise1, Leakage, Relax


def two_qubit_gate_leakage_noise_profile(pL: float) -> NoiseProfile:
    """
    Noise profile for leakage channels caused by
    two qubit gates.

    Parameters
    ----------
    pL: float
        Leakage channel probability parameter.

    Returns
    -------
    leakage_two_qubit_gate : NoiseProfile
        Leakage noise profile for two qubit gates.
    """

    def leakage_two_qubit_gate(noise_context):
        return [
            Leakage(qubit, pL)
            for qubit in noise_context.gate_layer_qubits(
                tuple(TWO_QUBIT_GATES), gate_qubit_count=2
            )
        ]


    return leakage_two_qubit_gate


def qubit_reset_leakage_noise_profile(pL: float) -> NoiseProfile:
    """
    Noise profile for leakage channels caused by
    reset gates.

    Parameters
    ----------
    pL: float
        Leakage channel probability parameter.

    Returns
    -------
    reset_leakage : NoiseProfile
        Leakage noise profile for reset gates.
    """

    def reset_leakage(noise_context):
        return [
            Leakage(qubit, pL)
            for qubit in noise_context.gate_layer_qubits(OneQubitResetGate)
        ]

    return reset_leakage


def two_qubit_gate_relaxation_noise_profile(pR: float) -> NoiseProfile:
    """
    Noise profile for relaxation channels caused by
    two qubit gates.

    Parameters
    ----------
    pR: float
        Relaxation channel probability parameter.

    Returns
    -------
    two_qubit_gate_relax : NoiseProfile
        Relaxation noise profile for two qubit gates.
    """

    def two_qubit_gate_relax(noise_context):
        return [
            Relax(qubit, pR)
            for qubit in noise_context.gate_layer_qubits(
                tuple(TWO_QUBIT_GATES), gate_qubit_count=2
            )
        ]

    return two_qubit_gate_relax


def one_qubit_clifford_gate_relaxation_noise_profile(pR: float) -> NoiseProfile:
    """
    Noise profile for relaxation channels caused by
    one qubit clifford gates.

    Parameters
    ----------
    pR: float
        Relaxation channel probability parameter.

    Returns
    -------
    one_qubit_relax : NoiseProfile
        Relaxation noise profile for ne qubit clifford gates.
    """

    def one_qubit_relax(noise_context):
        return [
            Relax(qubit, pR)
            for qubit in noise_context.gate_layer_qubits(
                (*tuple(TWO_QUBIT_GATES), OneQubitCliffordGate), gate_qubit_count=1
            )
        ]

    return one_qubit_relax


def idle_qubit_relaxation_noise_profile(pR: float) -> NoiseProfile:
    """
    Noise profile for relaxation channels on idle qubits.

    Parameters
    ----------
    pR: float
        Relaxation channel probability parameter.

    Returns
    -------
    idle_relax : NoiseProfile
        Relaxation noise profile for idle qubits.
    """

    def idle_relax(noise_context):
        idle_qubits = noise_context.circuit.qubits - set(
            noise_context.gate_layer_qubits(Gate)
        )
        return [Relax(qubit, pR) for qubit in idle_qubits]

    return idle_relax


def resonator_idle_qubit_relaxation_noise_profile(p: float, pR: float) -> NoiseProfile:
    """
    Noise profile for channels on resonator idle qubits. Adds
    relaxation and one qubit depolarising noise.

    Parameters
    ----------
    p: float
        One qubit depolarising channel probability parameter.
    pR: float
        Relaxation channel probability parameter.

    Returns
    -------
    resonator_idle : NoiseProfile
        Relaxation noise profile for idle qubits during active
        resonator operations.
    """
    depolarise1_generator_resonator_idle = Depolarise1.generator_from_prob(p)

    def resonator_idle(noise_context):
        noise_channels = []
        if resonator_qubits := set(
            noise_context.gate_layer_qubits(OneQubitMeasurementGate)
            + noise_context.gate_layer_qubits(OneQubitResetGate)
        ):
            idle_qubits = noise_context.circuit.qubits - resonator_qubits
            relaxations = [Relax(q, pR) for q in idle_qubits] if pR > 0 else []
            noise_channels = (
                depolarise1_generator_resonator_idle(idle_qubits) + relaxations
            )
        return noise_channels

    return resonator_idle



# (c) Copyright Riverlane 2020-2025.
"""`Simulation` module aggregates calls to different simulation backends.
"""

from __future__ import annotations

from pathlib import Path

import numpy as np
import stim

import deltakit_explorer
from deltakit_explorer.enums._api_enums import DataFormat
from deltakit_explorer.types._types import LeakageFlags, Measurements


def simulate_with_stim(
    stim_circuit: str | stim.Circuit,
    shots: int,
    result_file: str | Path | None = None,
    client: deltakit_explorer.Client | None = None,
) -> tuple[Measurements, LeakageFlags | None]:
    """This method simulates the circuit using
    a Clifford STIM simulator. Clifford simulation is fast,
    and supports large numbers of qubits, but is limited in
    the set of gates one may use, and noise mechanisms.

    Args:
        stim_circuit (str | stim.Circuit):
            Circuit in STIM language.
        shots (int):
            Number of shots.
        result_file (Optional[str | Path]):
            If provided, saves the data to this file. If None,
            saves them in RAM. This is useful, when you run
            large local simulations, and want to save measurements
            directly to the disk. Not supported for cloud calls.
        client (Optional[Client]):
            If provided, performs the simulation using the `client`.
            Required if `stim_circuit` includes leakage.
            Currently implemented only when `result_file` is ``None``.

    Returns:
        Tuple[Measurements, Optional[LeakageFlags]]:
            (Measurements, Leakage). Leakage may be None.


    Examples:
        Running a circuit 1000 times and saving to a `b8` file::

            meas = simulation.simulate_with_stim(
                stim_circuit=noisy_circuit,
                shots=1000,
                result_file="results.b8",
            )

        Running a circuit 1000 times and saving to RAM::

            meas = simulation.simulate_with_stim(
                stim_circuit=noisy_circuit,
                shots=1000,
            )

    """

    if shots < 0:
        msg = "Please specify a non-negative number of shots."
        raise ValueError(msg)

    if client is not None:
        circuit = str(stim_circuit.as_stim_circuit()) if not isinstance(stim_circuit, str) else stim_circuit
        if result_file is not None:
            msg = "Use of `client` is currently incompatible with `result_file`."
            raise NotImplementedError(msg)
        measurements, leakage_flags = client.simulate_stim_circuit(circuit, shots=shots)
        return measurements, leakage_flags

    circuit = stim.Circuit(stim_circuit) if isinstance(stim_circuit, str) else stim_circuit
    sampler = circuit.compile_sampler()
    # E.g. if we under 100MB, save to RAM
    # result_size = circuit.num_measurements * shots
    if result_file is None:
        result = sampler.sample(shots=shots).astype(np.uint8)
        return Measurements(result, data_format=DataFormat.B8), None
    data_format = DataFormat.B8
    sampler.sample_write(
        shots=shots,
        filepath=str(result_file),
        format=data_format.value,
    )
    return Measurements(
        Path(result_file),
        data_format=data_format,
        data_width=circuit.num_measurements
    ), None



# (c) Copyright Riverlane 2020-2025.
"""
This module contains standalone functions related to detectors used to generating
stabiliser code circuits.
"""
import itertools
import warnings
from collections import defaultdict
from collections.abc import Iterable as IterableC
from collections.abc import Sequence
from functools import reduce
from typing import TYPE_CHECKING

import numpy as np
from deltakit_circuit import (
    Circuit,
    Coordinate,
    Detector,
    MeasurementRecord,
    Qubit,
    ShiftCoordinates,
)

from deltakit_explorer.codes._stabiliser import Stabiliser

if TYPE_CHECKING:
    from deltakit_explorer.codes._css._css_stage import CSSStage


def _get_coordinate_from_ancilla_qubit(
    ancilla_qubit: Qubit | None,
) -> Coordinate | None:
    """
    Extracts and returns the coordinate of the qubit provided. If these coordinates
    are not available (i.e. if the qubit provided is None or if the qubit's unique
    identifier is not a set of coordinates), then returns None instead.

    Before returning the coordinates, initialises an extra coordinate dimension to 0,
    (representing time) such that the coordinate returned readily adopts a detector
    coordinate format.

    Parameters
    ----------
    ancilla_qubit : Qubit, optional
        The qubit of which to extract the coordinate. This function is meant for an
        ancilla qubit from a Stabiliser, which is why it handles the cases where the
        qubit is None or its unique identifier is not a Coordinate.

    Returns
    -------
    Coordinate | None
    """
    if ancilla_qubit is not None:
        unique_id = ancilla_qubit.unique_identifier
        qubit_coord = (
            tuple(unique_id) if isinstance(unique_id, IterableC) else (unique_id,)
        )
        if all(
            isinstance(coord, (float, int, np.floating, np.integer))
            for coord in qubit_coord
        ):
            full_coord = (*qubit_coord, 0)
            return Coordinate(*full_coord)

    return None


def _get_coordinate_from_data_qubits(
    data_qubits: list[Qubit],
) -> Coordinate | None:
    """
    Computes and returns a coordinate which is the average in each coordinate
    dimension of qubits' coordinates provided. If any of these qubits' coordinate is
    not valid (i.e. if any unique identifier is not a set of coordinates), then
    returns None instead.

    Before returning the coordinates, initialises an extra coordinate dimension to 0,
    (representing time) such that the coordinate returned readily adopts a detector
    coordinate format.

    Parameters
    ----------
    data_qubits : List[Qubit], optional
        The qubits from which to compute the coordinate. This function is meant for
        data qubits from a Stabiliser, which is why it handles the cases where the
        qubits' unique identifiers are not Coordinates.

    Returns
    -------
    Coordinate | None
    """
    qubit_coords = [
        (
            tuple(qubit.unique_identifier)
            if isinstance(qubit.unique_identifier, IterableC)
            else (qubit.unique_identifier,)
        )
        for qubit in data_qubits
    ]

    # if any elements are not floats or ints, a valid coordinate cannot be
    # constructed
    if any(
        any(not isinstance(coord, (float, int)) for coord in qubit_coord)
        for qubit_coord in qubit_coords
    ):
        return None

    # if the data qubit coordinates are not all of the same length, a valid
    # detector coordinate cannot be constructed
    if any(
        len(qubit_coord) != len(qubit_coords[0]) for qubit_coord in qubit_coords[1:]
    ):
        return None
    qubit_coord_len = len(qubit_coords[0])

    full_coord = (
        *tuple(
            float(np.mean([qubit_coord[i] for qubit_coord in qubit_coords]))
            for i in range(qubit_coord_len)
        ), 0)
    return Coordinate(*full_coord)


def _calculate_detector_coordinates(
    stabilisers: Sequence[Stabiliser],
) -> tuple[Coordinate, ...]:
    """
    Calculate detector coordinates based on the stabilisers used in the stage.

    Where possible, the coordinate will be taken from the ancilla qubit unique
    identifier. If this is not possible, an attempt will be made to form the
    coordinate from averaging the data qubit unique identifiers.

    If, due to averaging the data qubit unique identifier, multiple
    coordinates are equal, small perturbations are applied.

    However, if it is not possible to form coordinates of the same dimension from
    all stabilisers, or there are further clashes after this process, the coordinates
    will be formed by indexing the stabilisers by their order of input.

    Parameters
    ----------
    stabilisers : Sequence[Stabiliser]
        The stabilisers from which we wish to calculate detector coordinates.

    Returns
    -------
    Tuple[Coordinate, ...]
        Detector coordinates for the input stabilisers, with the ordering the same
        as in the input.
    """

    detector_coords: list[Coordinate] = []
    coord_len: int = -1
    for stabiliser in stabilisers:
        # attempt to get a coordinate from the ancilla qubit
        detector_coord = _get_coordinate_from_ancilla_qubit(stabiliser.ancilla_qubit)
        if (
            detector_coord is not None
            # check if detector coordinate length is consistent with previous lengths
            and (len(detector_coords) == 0 or coord_len == len(detector_coord))
        ):
            detector_coords.append(detector_coord)
            coord_len = len(detector_coord)
            continue

        # attempt to get a coordinate from the data qubits
        detector_coord = _get_coordinate_from_data_qubits(
            [pauli.qubit for pauli in stabiliser.paulis if pauli is not None]
        )
        if detector_coord is None:
            break

        # check if detector coordinate length is consistent with previous lengths
        if len(detector_coords) != 0 and len(detector_coord) != coord_len:
            break

        detector_coords.append(detector_coord)
        coord_len = len(detector_coord)

    else:
        # check for repeats
        for detector_coord in detector_coords:
            if (detector_count := detector_coords.count(detector_coord)) > 1:
                indices = [
                    i
                    for i in range(len(detector_coords))
                    if detector_coords[i] == detector_coord
                ]
                # spread repeated coordinates along their first index
                for i_index, index in enumerate(indices):
                    first_coord = (
                        detector_coord[0] - 0.3 + 0.6 * i_index / (detector_count - 1)
                    )
                    full_coord = (first_coord, *tuple(coord for coord in detector_coord[1:]))
                    detector_coords[index] = Coordinate(*full_coord)

        # check again for repeats as some shifted coordinates with different original
        # coordinates may now overlap
        for detector_coord in detector_coords:
            if detector_coords.count(detector_coord) > 1:
                break
        else:
            return tuple(detector_coords)

    return tuple(Coordinate(i, 0) for i in range(len(stabilisers)))


def _get_coordinate_shifts(detectors: list[Detector]) -> list[ShiftCoordinates]:
    """
    Given a list of detectors, computes coordinate shifts which increase the last
    coordinate dimension by 1. If the detectors provided have coordinates of
    different number of dimensions, then a coordinate shift for each is added.
    A warning is raised in that case.

    Parameters
    ----------
    detectors : List[Detector]
        The list of detectors to use to generate the shift coordinate instructions.

    Returns
    -------
    List[ShiftCoordinates]
        A list of the shift coordinate instructions.
    """

    detector_num_dimensions = [
        len(detector.coordinate)
        for detector in detectors
        if detector.coordinate is not None
    ]

    # obtaining the unique elements in this way to conserve the order
    unique_num_dimensions = list(dict.fromkeys(detector_num_dimensions))

    if len(unique_num_dimensions) > 1:
        warnings.warn(
            "Detector coordinates have different dimensions and thus "
            + "coordinates may shift in multiple directions."
        )

    return [
        ShiftCoordinates((0,) * (num_dimensions - 1) + (1,))
        for num_dimensions in unique_num_dimensions
    ]


def _get_between_round_detectors(
    detector_coordinates: Sequence[Coordinate],
    num_additional_measurements: int = 0,
) -> list[Detector]:
    """
    For a scenario where multiple rounds of syndrome extraction of a same group of
    stabilisers are performed, compute and return the detectors which compare
    measurements lying in different rounds.

    Each detector returned is associated with the measurement (or measurement pair
    across rounds) of the ancilla qubit of one of the stabilisers. We therefore
    obtain one detector per stabiliser. Our detectors consist of measurement pairs
    from two adjacent rounds.
    Parameters
    ----------
    detector_coordinates : Sequence[Coordinate, ...]
        A list of the detector coordinates associated with each of the stabilisers
        measured in each syndrome extraction round, ordered in the same way.
    num_additional_measurements : int, optional
        The number of additional measurements taking place at the end of each round of
        syndrome extraction.

    Returns
    -------
    List[Detector]
        Between-round detectors for the stabilisers.
    """
    if len(detector_coordinates) == 0:
        return []

    num_measurements = len(detector_coordinates) + num_additional_measurements

    detectors: list[Detector | ShiftCoordinates] = []

    for i, detector_coord in enumerate(detector_coordinates):
        lookback_index = i - num_measurements
        detector_mmts = [MeasurementRecord(lookback_index)]

        detector_mmts.append(MeasurementRecord(lookback_index - num_measurements))

        detectors.append(Detector(detector_mmts, detector_coord))

    return detectors


def get_between_round_detectors_and_coordinate_shifts(
    detector_coordinates: Sequence[Coordinate],
    num_additional_measurements: int = 0,
) -> list[Detector | ShiftCoordinates]:
    """
    For a scenario where multiple rounds of syndrome extraction of a same group of
    stabilisers are performed, compute and return the detectors which compare
    measurements lying in different rounds.

    Each detector returned is associated with the measurement (or measurement pair
    across rounds) of the ancilla qubit of one of the stabilisers. We therefore
    obtain one detector per stabiliser. Our detectors consist of measurement pairs
    from two adjacent rounds.

    Note a ShiftCoordinate instruction which increases the last detector coordinate
    dimension (i.e. time) by 1 is appended to the end of the detectors list returned.

    Parameters
    ----------
    detector_coordinates : Sequence[Coordinate]
        A list of the detector coordinates associated with each of the stabilisers
        measured in each syndrome extraction round, ordered in the same way.
    num_additional_measurements : int, optional
        The number of additional measurements taking place at the end of each round of
        syndrome extraction.

    Returns
    -------
    List[Union[Detector, ShiftCoordinates]]
        Between-round detectors for the stabilisers along with a ShiftCoordinate instruction
        that increases the last detector coordinate dimension.
    """
    detectors = _get_between_round_detectors(
        detector_coordinates, num_additional_measurements
    )
    shifts = _get_coordinate_shifts(detectors)

    return detectors + shifts


def _get_joint_sub_super_stabilisers_ind(
    previous_stabilisers: Sequence[Stabiliser],
    current_stabilisers: Sequence[Stabiliser],
) -> list[tuple[tuple[int, ...], tuple[int, ...]]]:
    """
    Exhaustively search for one element in previous_stabilisers/current_stabilisers and
    one or more elements from current_stabilisers/previous_stabilisers such that the
    product of the latter is the former. The method collects the indices of these
    elements, e.g.
        - ((i_prev,),(i_curr,j_curr,...)) corresponds to one element of
            previous_stabilisers that can be obtained as a product of elements from
            current_stabilisers, or
        - ((i_prev,j_prev,...),(i_curr)) corresponds to one element of
            current_stabilisers that can be obtained as a product of elements from
            previous_stabilisers,

    Parameters
    ----------
    previous_stabilisers : Sequence[Stabiliser]
        Pauli products present in previous CSSStage.
    current_stabilisers : Sequence[Stabiliser]
        Pauli products present in current CSSStage.

    Returns
    -------
    List[Tuple[Tuple[int, ...], Tuple[int, ...]]]
        The list collecting the stabiliser outcome comparisons.
    """
    # First build dictionaries of the comparisons
    joint_stabilisers_ind: dict[int, int] = {}
    sub_stabilisers_ind: dict[int, list[int]] = defaultdict(list)
    super_stabilisers_ind: dict[int, list[int]] = defaultdict(list)

    for (ind_prev, stab_prev), (ind_curr, stab_curr) in itertools.product(
        enumerate(previous_stabilisers), enumerate(current_stabilisers)
    ):
        if stab_prev.operator_repr == stab_curr.operator_repr:
            joint_stabilisers_ind[ind_prev] = ind_curr
        elif stab_prev.operator_repr < stab_curr.operator_repr:
            sub_stabilisers_ind[ind_curr].append(ind_prev)
        elif stab_prev.operator_repr > stab_curr.operator_repr:
            super_stabilisers_ind[ind_prev].append(ind_curr)

    # Check if the small Paulis multiply together to the big Pauli.
    for ind_curr in list(sub_stabilisers_ind.keys()):
        indices_prev = sub_stabilisers_ind[ind_curr]
        previous_stabiliser_product = reduce(
            lambda x, y: x * y,
            [previous_stabilisers[ind_prev] for ind_prev in indices_prev],
        )
        if (
            current_stabilisers[ind_curr].operator_repr != previous_stabiliser_product.operator_repr
        ):
            del sub_stabilisers_ind[ind_curr]

    for ind_prev in list(super_stabilisers_ind.keys()):
        indices_curr = super_stabilisers_ind[ind_prev]
        current_stabiliser_product = reduce(
            lambda x, y: x * y,
            [current_stabilisers[ind_curr] for ind_curr in indices_curr],
        )
        if (
            previous_stabilisers[ind_prev].operator_repr != current_stabiliser_product.operator_repr
        ):
            del super_stabilisers_ind[ind_prev]
    # Convert these dictionaries into the list format
    # fmt: off
    full_index_list: list[tuple[tuple[int, ...], tuple[int, ...]]] = (
        [
            ((ind_prev,), (ind_curr,))
            for ind_prev, ind_curr in sorted(
                joint_stabilisers_ind.items(), key=lambda item: item[0]
            )
        ]
        + [
            (tuple(indices_prev), (ind_curr,))
            for ind_curr, indices_prev in sorted(
                sub_stabilisers_ind.items(), key=lambda item: item[0]
            )
        ]
        + [
            ((ind_prev,), tuple(indices_curr,))
            for ind_prev, indices_curr in sorted(
                super_stabilisers_ind.items(), key=lambda item: item[0]
            )
        ]
    )
    # fmt: on
    return full_index_list


def _get_stage_transition_detectors(
    previous_stage: "CSSStage", current_stage: "CSSStage"
) -> list[Detector]:
    """
    Return a list containing all detectors for transitioning between the
    stages previous_stage and current_stage.

    Parameters
    ----------
    previous_stage : CSSStage
        A CSSStage from which we transition. For instance, for patch growing
        previous_stage would be the stage where we measure a small patch for some
        rounds.
    current_stage : CSSStage
        The CSSStage into which we intend to transition. For instance, for patch
        growing current_stage would be the stage where we measure the bigger patch for
        some rounds.

    Returns
    -------
    Circuit
        The transition detectors.
    """
    detectors = []
    previous_stabilisers = (
        previous_stage.ordered_stabilisers + previous_stage.resets_as_stabilisers
    )
    current_stabilisers = (
        current_stage.measurements_as_stabilisers + current_stage.ordered_stabilisers
    )
    current_stabilisers_before = (
        current_stage.measurements_as_stabilisers + current_stage.stabilisers_before
    )

    # Do comparison with current_stabilisers_before, e.g. if current_stage does a
    # transversal H or SWAP, we need to compare with the stabilisers before the
    # transversal gates.
    joint_sub_super_stabilisers_ind = _get_joint_sub_super_stabilisers_ind(
        previous_stabilisers=previous_stabilisers,
        current_stabilisers=current_stabilisers_before,
    )

    current_offset = len(current_stabilisers_before)
    previous_offset = current_offset + len(previous_stage.ordered_stabilisers)

    # Collect detectors
    for indices_prev, indices_curr in joint_sub_super_stabilisers_ind:
        # Add indices for current stabiliser measurement
        rec_indices = [
            MeasurementRecord(ind_curr - current_offset) for ind_curr in indices_curr
        ]

        # Add indices from previous stabiliser measurement, ignoring resets
        rec_indices += [
            MeasurementRecord(ind_prev - previous_offset)
            for ind_prev in indices_prev
            if ind_prev < len(previous_stage.ordered_stabilisers)
        ]

        rec_indices.sort(key=lambda meas_rec: meas_rec.lookback_index, reverse=True)

        # Determine if "big" stabiliser is from previous or current stage
        big_stab_from_current_stage = len(indices_curr) == 1

        # Get the "big" stabiliser
        big_stabiliser = (
            current_stabilisers[indices_curr[0]]
            if big_stab_from_current_stage
            else previous_stabilisers[indices_prev[0]]
        )
        # Get the detector coordinate
        stage = current_stage if big_stab_from_current_stage else previous_stage
        stab_index = stage.ordered_stabilisers.index(big_stabiliser)
        detector_coordinate = stage.detector_coordinates[stab_index]

        detectors.append(
            Detector(
                measurements=rec_indices,
                coordinate=detector_coordinate,
            )
        )

    return detectors


def get_stage_transition_circuit(
    previous_stage: "CSSStage", current_stage: "CSSStage"
) -> Circuit:
    """
    Return a Circuit containing all the DETECTOR and SHIFT_COORD information for
    transitioning between previous_stage and current_stage stages.

    Parameters
    ----------
    previous_stage : CSSStage
        A CSSStage from which we transition. For instance, for patch growing
        previous_stage would be the stage where we measure a small patch for some
        rounds.
    current_stage : CSSStage
        The CSSStage into which we intend to transition. For instance, for patch
        growing current_stage would be the stage where we measure the bigger patch for
        some rounds.

    Returns
    -------
    Circuit
        The detectors and the coordinate shift instruction.
    """
    detectors = _get_stage_transition_detectors(previous_stage, current_stage)
    shifts = _get_coordinate_shifts(detectors)

    return Circuit(detectors + shifts)



# (c) Copyright Riverlane 2020-2025.
"""This file contains implementation of an API for accessing
the Deltakit service.
"""

from __future__ import annotations

import json
from typing import TYPE_CHECKING, Any, cast
from urllib.parse import urljoin

import numpy as np
import numpy.typing as npt
import requests
import requests.adapters
from typing_extensions import override

if TYPE_CHECKING:
    import stim
from gql import Client, gql
from gql.client import SyncClientSession
from gql.transport.exceptions import TransportQueryError
from gql.transport.requests import RequestsHTTPTransport
from graphql import ExecutionResult

from deltakit_explorer._api._api_client import APIClient
from deltakit_explorer._api._auth import (
    get_token,
    https_verification_disabled,
    set_token,
)
from deltakit_explorer._utils._logging import Logging
from deltakit_explorer.enums._api_enums import APIEndpoints, DataFormat
from deltakit_explorer.types import (
    DataString,
    Decoder,
    DecodingResult,
    DetectionEvents,
    LeakageFlags,
    Measurements,
    NoiseModel,
    ObservableFlips,
    QubitCoordinateToDetectorMapping,
)
from deltakit_explorer.types._exceptions import ServerException
from deltakit_explorer.types._experiment_types import QECExperimentDefinition


# pylint: disable=too-many-instance-attributes
class GQLClient(APIClient):
    """This is a parent class which contains methods
    to GQL server functionality.
    """

    def __init__(
        self,
        base_url: str,
        request_timeout: int = 60000,
    ):
        """
        The constructor accepts and endpoint and a timeout.
        `base_url` is a URL of the service.
        - `/api/graphql/` accepts GraphQL POST-queries,
        - `/api/data/` allows GET-requests of content.

        Args:
            base_url (str): Service URL.
            request_timeout (int):
                request timeout (seconds). Default is 60_000 seconds.

        Examples:
            This is how an instance of the child class is created::

                client = Client(
                    base_url="http://deltakit.rivelane.com/proxy"
                )

        """
        super().__init__(base_url, request_timeout)
        self.content_endpoint = f"{self.base_url}/api/data/"
        self.graphql_endpoint = f"{self.base_url}/api/graphql"

        self.auth_headers: dict[str, str | bytes] = {}
        self.transport = self._get_transport()
        self.client = self._get_client()
        self._session: SyncClientSession = self.client.connect_sync()
        self._request_session = requests.Session()
        retries = requests.adapters.Retry(
                total=3, backoff_factor=0.5, status_forcelist=[500, 502, 503, 504],
        )
        self._request_session.mount(
            "http://",
            requests.adapters.HTTPAdapter(max_retries=retries),
        )
        self._request_session.mount(
            "https://",
            requests.adapters.HTTPAdapter(max_retries=retries),
        )
        self._request_session.headers = self.auth_headers

    def _update_headers(self):
        self.auth_headers = {
            "Authorization": "Bearer " + get_token()
        }

    def _get_transport(self):
        self._update_headers()
        return RequestsHTTPTransport(
            url=self.graphql_endpoint,
            headers=self.auth_headers,
            timeout=self.request_timeout,
            verify=not https_verification_disabled(),
            stream=True,
        )

    def _get_client(self):
        return Client(
            transport=self.transport,
            fetch_schema_from_transport=False,
            execute_timeout=self.request_timeout,
        )

    @staticmethod
    def _get_message(code: int) -> str:
        if code == 503:
            return "Service unavailable"
        if code == 404:
            return "URL not found"
        return ""

    def _get_query(self, name: str, request_id: str) -> str:
        """Downloads a GraphQL query given a name,
        and substitutes provided named arguments.

        Args:
            name (str): query name.
            request_id (str): request_id for tracking in logs.

        Returns:
            str: GraphQL parametric query.
        """
        url = (
            urljoin(self.content_endpoint + "/query/", name)
            + f"?request_id={request_id}"
        )
        resp = self._request_session.get(
            url,
            headers=self.auth_headers,
            verify=not https_verification_disabled(),
            timeout=self.client.execute_timeout,
        )
        if resp.status_code != 200:
            # gateway responds in json,
            # but it may also forward errors from the server
            try:
                error_dict = json.loads(resp.text)
                if isinstance(error_dict, dict):
                    # HTTP error from the app
                    server_message = error_dict.get("detail", resp.text)
                    # Auth error from the gateway
                    gateway_message = error_dict.get("message", server_message)
                    error_code = error_dict.get("error_code", resp.status_code)
                else:
                    # is not a dict
                    error_code = resp.status_code
                    gateway_message = resp.text
            except json.JSONDecodeError:
                # could not parse as JSON
                gateway_message = resp.text
                error_code = resp.status_code
            if not gateway_message:
                gateway_message = self._get_message(error_code)
            # probably, html content
            if len(gateway_message) > 160:
                gateway_message = gateway_message[:160] + "..."
            msg = f"Status {resp.status_code} (Error #{error_code}): {gateway_message}"
            raise ServerException(
                msg
            )
        return resp.text

    def execute_query(
        self, query: str, variable_values: dict, query_id: str
    ) -> dict[str, Any]:
        """Executes arbitrary GraphQL query and
        returns server response as a python dicr.

        Args:
            query (str): GraphQL query text.
            variable_values (Dict): arguments.
            query_id (str): identifier of the request.

        Returns:
            Dict[str, Any]: Server response.

        Raises:
            TransportQueryError
        """
        request = gql(query)
        try:
            result = self._session.execute(request, variable_values=variable_values)
            if isinstance(result, ExecutionResult) and result.errors:
                raise ServerException(str(result.errors[0].message))
            return cast(dict[str, Any], result)
        except TransportQueryError as err:
            Logging.error(err, query_id)
            message = str(err.errors[0].get("message") if err.errors else err)
            raise ServerException(message) from err
        except KeyboardInterrupt:
            count = self.kill(query_id)
            Logging.warn(f"Cancelled a task, {count} worker(s) cleared.", query_id)
            raise

    @override
    def execute(
        self,
        query_name: APIEndpoints,
        variable_values: dict,
        request_id: str,
    ) -> dict[str, Any]:
        query = self._get_query(query_name.value, request_id)
        return self.execute_query(query, variable_values, request_id)

    def kill(self, request_id: str) -> int:
        resp = self._request_session.get(
            f"{self.content_endpoint}kill/{request_id}",
            headers=self.auth_headers,
            verify=not https_verification_disabled(),
            timeout=10,
        )
        return int(resp.text)

    @override
    def set_token(self, token: str, validate: bool = True):
        try:
            old_token = get_token()
        except RuntimeError as err:  # pragma: nocover
            Logging.warn(f"No token was set previously. {err.args[0]}", "set_token")
            old_token = ""  # nosec B105

        set_token(token)
        self._update_headers()
        self._request_session.headers = self.auth_headers
        if validate:
            try:
                # test a content endpoint
                self._get_query("decode", "set_token")
            except ServerException as exc:
                Logging.error(exc, "set_token")
                set_token(old_token)
                msg = f"Token failed validation: {exc.message}."
                raise ServerException(
                    msg
                ) from exc
            except requests.exceptions.ConnectionError as exc_coonection:
                message = "Could not validate token: cannot reach client."
                set_token(old_token)
                Logging.error(exc_coonection, "set_token")
                raise ServerException(message) from exc_coonection
        msg = (
            "Token successfully validated and "
            "will be used automatically in future sessions."
        )
        Logging.info(msg, "set_token")

    @override
    def generate_circuit(self, experiment_definition: QECExperimentDefinition, request_id: str) -> str:
        result = self.execute(  # pragma: nocover
            query_name=APIEndpoints.GENERATE_CIRCUIT,
            variable_values={
                "expType": experiment_definition.experiment_type.value,
                "codeType": experiment_definition.code_type.value,
                "observableBasis":
                    experiment_definition.observable_basis.value,
                "basisGates": experiment_definition.basis_gates,
                "rounds": experiment_definition.num_rounds,
                "outputCircuitLocation": "",
                "outputCircuitName": DataString.empty,
                "generateCircuitParameters":
                    experiment_definition.get_parameters_gql_string(),
                "requestId": request_id,
            },
            request_id=request_id,
        )
        return DataString.from_data_string(
            result["generateCircuit"]["uid"]).to_string()

    @override
    def simulate_circuit(self, stim_circuit: str | stim.Circuit, shots: int, request_id: str) -> tuple[Measurements, LeakageFlags | None]:
        output_format = DataFormat.F01
        result = self.execute(
            query_name=APIEndpoints.SIMULATE_CIRCUIT,
            variable_values={
                "circuit": str(DataString(str(stim_circuit))),
                "location": "",
                "leakageLocation": "",
                "outputData": DataString.empty,
                "outputLeakageData": DataString.empty,
                "outputFormat": output_format.value,
                "outputLeakageFormat": output_format.value,
                "shots": shots,
                "requestId": request_id,
            },
            request_id=request_id,
        )
        uids = [
            DataString.from_data_string(uri["uid"])
            for uri in result["simulateWithStim"]
        ]
        if len(uids) == 1:  # no leakage returned
            return (Measurements(uids[0], data_format=output_format), None)
        if len(uids) == 2:
            return (
                Measurements(uids[0], data_format=output_format),
                LeakageFlags(uids[1], data_format=output_format),
            )
        msg = f"Expected 1 or 2 result files, but got {len(uids)}"
        raise ServerException(msg)

    @override
    def add_noise(self, stim_circuit: str | stim.Circuit, noise_model: NoiseModel, request_id: str) -> str:
        variables = {
            "circuit": str(DataString(str(stim_circuit))),
            "result_file": DataString.empty,
            "requestId": request_id,
            "result_folder": "",
            **noise_model.__dict__,
        }
        if noise_model.ENDPOINT is None:
            msg = f"Noise addition for {type(noise_model)} is not implemented."
            raise NotImplementedError(msg)
        result = self.execute(
            query_name=noise_model.ENDPOINT, variable_values=variables, request_id=request_id)
        dstring = DataString.from_data_string(result[noise_model.ENDPOINT_RESULT_FIELDNAME]["uid"])
        return dstring.to_string()

    @override
    def decode(self, detectors: DetectionEvents,
        observables: ObservableFlips,
        decoder: Decoder,
        noisy_stim_circuit: str | stim.Circuit,
        leakage_flags: LeakageFlags | None,
        request_id: str,
    ) -> DecodingResult:
        parameters = []
        if decoder.parameters is not None:
            for key, value in decoder.parameters.items():
                parameters.append([key, str(value)])
        use_format = DataFormat.B8
        # used to extract size
        predictions_format = DataFormat.F01
        variable_values = {
            "contentType": use_format.value,
            "decoder": decoder.decoder_type.value,
            "detFile": detectors.as_data_string(use_format),
            "obsFile": observables.as_data_string(use_format),
            "useExperimentalGraph": decoder.use_experimental_graph,
            "noisyStimFile": str(DataString(str(noisy_stim_circuit))),
            "parameters": parameters,
            "requestId": request_id,
            "resultFileLocation": "",
            "resultFile": DataString.empty,
            "resultContentType": predictions_format.value,
            "threads": decoder.parallel_jobs,
        }
        if leakage_flags is not None:
            query_name = APIEndpoints.DECODE_LEAKAGE
            variable_values["leakageFile"] = leakage_flags.as_data_string(use_format)
        else:
            query_name = APIEndpoints.DECODE
        Logging.info(
            f"Decoding request {request_id} has been sent to a client.",
            request_id
        )
        result = self.execute(query_name, variable_values, request_id)
        return DecodingResult(
            predictions_format=predictions_format,
            **result["decode"],
        )

    @override
    def defect_rates(
        self,
        detectors: DetectionEvents,
        stim_circuit: str | stim.Circuit,
        request_id: str,
    ) -> dict[tuple[float, ...], list[float]]:
        use_format = DataFormat.B8
        result = self.execute(
            query_name=APIEndpoints.DEFECT_RATES,
            variable_values={
                "contentType": use_format.value,
                "detFile": detectors.as_data_string(use_format),
                "stimFile": str(DataString(str(stim_circuit))),
                "coordinates": None,
                "requestId": request_id,
            },
            request_id=request_id,
        )
        rates = result["defectRates"]
        return {
            tuple(map(float, pair["key"])): pair["value"]
            for pair in rates["items"]
        }

    @override
    def get_correlation_matrix_for_trimmed_data(
        self,
        detectors: DetectionEvents,
        noise_floor_circuit: str | stim.Circuit,
        use_default_noise_model_edges: bool,
        request_id: str,
    ) -> tuple[npt.NDArray[np.float64], QubitCoordinateToDetectorMapping]:
        use_format = DataFormat.B8
        result = self.execute(
            query_name=APIEndpoints.CORRELATION_MATRIX,
            variable_values={
                "contentType": use_format.value,
                "detFile": detectors.as_data_string(use_format),
                "stimFile": str(DataString(str(noise_floor_circuit))),
                "useStimGraph": use_default_noise_model_edges,
                "requestId": request_id,
            },
            request_id=request_id,
        )
        matrix = result["correlationMatrix"]["matrix"]
        qubit_to_detectors = result["correlationMatrix"]["qubitToDetectionEvents"]
        return np.array(matrix), QubitCoordinateToDetectorMapping({
            tuple(qubit_det["qubit"]): qubit_det["detectors"] for qubit_det in qubit_to_detectors
        })

    @override
    def trim_circuit_and_detectors(
        self,
        stim_circuit: str | stim.Circuit,
        detectors: DetectionEvents,
        request_id: str
    ) -> tuple[str, DetectionEvents]:
        use_format = DataFormat.B8
        result = self.execute(
            query_name=APIEndpoints.TRIM_CIRCUIT_AND_DETECTORS,
            variable_values={
                "contentType": detectors.data.data_format.value,
                "data": detectors.as_data_string(use_format),
                "circuit": str(DataString(str(stim_circuit))),
                "trimmedCircuit": DataString.empty,
                "trimmedDetectorData": DataString.empty,
                "location": "",
                "outputFormat": use_format.value,
                "requestId": request_id,
            },
            request_id=request_id,
        )
        response = result["trimCircuitAndDetectionEvents"]
        return (
            DataString.from_data_string(response["circuit"]["uid"]).to_string(),
            DetectionEvents(
                data=DataString.from_data_string(response["detectors"]["uid"]),
                data_format=use_format,
            )
        )



# (c) Copyright Riverlane 2020-2025.
"""
This module contains classes used to capture native gate sets and gate times
of a QPU.
"""

from typing import Optional

from deltakit_circuit.gates import (
    MEASUREMENT_GATES,
    ONE_QUBIT_GATES,
    ONE_QUBIT_MEASUREMENT_GATES,
    RESET_GATES,
    TWO_QUBIT_GATES,
    Gate,
    OneQubitCliffordGate,
    OneQubitResetGate,
    TwoOperandGate,
    _MeasurementGate,
)
from deltakit_circuit.gates._measurement_gates import MPP

from deltakit_explorer._gates._gate_sets import (
    DEFAULT_MEASUREMENT_GATES,
    DEFAULT_ONE_QUBIT_GATES,
    DEFAULT_RESET_GATES,
    DEFAULT_TWO_QUBIT_GATES,
)


class NativeGateSetAndTimes:
    """
    Class for capturing native gate sets of a quantum computer and times of
    gate execution (in seconds).

    Parameters
    ----------
    one_qubit_gates : Optional[Dict[Type[OneQubitCliffordGate], float]]
        Dictionary of the one-qubit gates available on the quantum computer and the
        associated gate times. By default, the gates are those in
        DEFAULT_ONE_QUBIT_GATES with times all equal to 1.0.
    two_qubit_gates : Optional[Dict[Type[TwoOperandGate], float]]
        Dictionary of the two-qubit gates available on the quantum computer and the
        associated gate times. By default, the gates are those in
        DEFAULT_TWO_QUBIT_GATES with times all equal to 1.0.
    reset_gates : Optional[Dict[Type[OneQubitResetGate], float]]
        Dictionary of the reset gates available on the quantum computer and the
        associated gate times. By default, the gates are those in DEFAULT_RESET_GATES
        with times all equal to 1.0.
    measurement_gates : Optional[Dict[Type[_MeasurementGate], float]]
        Dictionary of the one-qubit measurement gates available on the quantum
        computer and the associated gate times. By default, the gates are those in
        DEFAULT_MEASUREMENT_GATES with times all equal to 1.0.

    Raises
    ------
    ValueError
        If any supplied one-qubit gate is not a valid one-qubit gate.
    ValueError
        If any supplied two-qubit gate is not a valid two-qubit gate.
    ValueError
        If any supplied reset gate is not a valid reset gate.
    ValueError
        If any supplied measurement gate is not a valid one-qubit measurement gate.
    ValueError
    ValueError
        If any time is not a positive float.
    """

    def __init__(
        self,
        one_qubit_gates: dict[type[OneQubitCliffordGate], float] | None = None,
        two_qubit_gates: dict[type[TwoOperandGate], float] | None = None,
        reset_gates: dict[type[OneQubitResetGate], float] | None = None,
        measurement_gates: dict[type[_MeasurementGate], float] | None = None,
    ):
        self.one_qubit_gates = (
            dict.fromkeys(DEFAULT_ONE_QUBIT_GATES, 1.0)
            if one_qubit_gates is None
            else one_qubit_gates
        )
        self.two_qubit_gates = (
            dict.fromkeys(DEFAULT_TWO_QUBIT_GATES, 1.0)
            if two_qubit_gates is None
            else two_qubit_gates
        )
        self.reset_gates = (
            dict.fromkeys(DEFAULT_RESET_GATES, 1.0)
            if reset_gates is None
            else reset_gates
        )
        self.measurement_gates = (
            dict.fromkeys(DEFAULT_MEASUREMENT_GATES, 1.0)
            if measurement_gates is None
            else measurement_gates
        )

        self.native_gates = (
            set(self.one_qubit_gates)
            | set(self.two_qubit_gates)
            | set(self.reset_gates)
            | set(self.measurement_gates)
        )

        if not set(self.one_qubit_gates) <= ONE_QUBIT_GATES:
            msg = "Element in one-qubit gate list is not a valid gate."
            raise ValueError(msg)

        if not set(self.two_qubit_gates) <= TWO_QUBIT_GATES:
            msg = "Element in two-qubit gate list is not a valid gate."
            raise ValueError(msg)

        if not set(self.reset_gates) <= RESET_GATES:
            msg = "Element in reset gate list is not a valid gate."
            raise ValueError(msg)

        if not set(self.measurement_gates) <= ONE_QUBIT_MEASUREMENT_GATES.union({MPP}):
            msg = "Element in measurement gate list is not a valid gate."
            raise ValueError(msg)

        gate_time_dicts = [
            self.one_qubit_gates,
            self.two_qubit_gates,
            self.reset_gates,
            self.measurement_gates,
        ]
        for gate_time_dict in gate_time_dicts:
            for gate, time in gate_time_dict.items():
                self._check_time(gate, time)

    @staticmethod
    def _check_time(gate: type[Gate], time: float) -> None:
        if time < 0.0:
            msg = (
                "A gate time must be a non-negative float but that for "
                f"{gate.stim_string} is {time}."
            )
            raise ValueError(msg)

    def add_gate(self, gate: type[Gate], time: float = 1.0) -> None:
        """
        Add a gate and associated time to the native gate set.

        Parameters
        ----------
        gate : Type[Gate]
            Gate to be added.
        time : float, optional
            Time of the gate to be added. By default, 1.0.
        """
        self._check_time(gate, time)
        if gate in ONE_QUBIT_GATES:
            self.one_qubit_gates[gate] = time
        elif gate in TWO_QUBIT_GATES:
            self.two_qubit_gates[gate] = time
        elif gate in RESET_GATES:
            self.reset_gates[gate] = time
        elif gate in ONE_QUBIT_MEASUREMENT_GATES.union({MPP}):
            self.measurement_gates[gate] = time
        else:
            msg = f"Unknown gate {gate} supplied."
            raise ValueError(msg)
        self.native_gates.add(gate)

    @staticmethod
    def from_times(time_1_qubit_gate: float, time_2_qubit_gate: float,
                   time_reset: float, time_measurement: float,
                   native_gates: Optional['NativeGateSet'] = None):
        """Assign times to gates based on class (1-qubit, 2-qubit, reset, measurement).

        Parameters
        ----------
        time_1_qubit_gate: float
            Time to execute a 1-qubit gate.
        time_2_qubit_gate: float
            Time to execute a 2-qubit gate.
        time_measurement: float
            Time to measure a qubit.
        time_reset: float
            Time to reset a qubit.
        native_gates: Optional[NativeGateSet]
            An instance of `NativeGateSet` specifying the native gates of the QPU.

        Returns
        -------
        native_gates_and_times: NativeGateSetAndTimes
            A `NativeGateSetAndTimes` object representing the native gates and times
            of a QPU.

        """

        if native_gates is not None:
            one_qubit_gates = native_gates.one_qubit_gates
            two_qubit_gates = native_gates.two_qubit_gates
            reset_gates = native_gates.reset_gates
            measurement_gates = native_gates.measurement_gates
        else:
            one_qubit_gates = ONE_QUBIT_GATES
            two_qubit_gates = TWO_QUBIT_GATES
            reset_gates = RESET_GATES
            measurement_gates = MEASUREMENT_GATES

        return NativeGateSetAndTimes(
            one_qubit_gates=dict.fromkeys(one_qubit_gates, time_1_qubit_gate),
            two_qubit_gates=dict.fromkeys(two_qubit_gates, time_2_qubit_gate),
            reset_gates=dict.fromkeys(reset_gates, time_reset),
            measurement_gates=dict.fromkeys(measurement_gates, time_measurement),
        )


class NativeGateSet(NativeGateSetAndTimes):
    """
    Class for capturing native gate sets of a quantum computer.

    Parameters
    ----------
    one_qubit_gates : Optional[Set[Type[OneQubitCliffordGate]]]
        Set of the one-qubit gates available on the quantum computer. By default,
        the gates are those in DEFAULT_ONE_QUBIT_GATES.
    two_qubit_gates : Optional[Set[Type[TwoOperandGate]]]
        Set of the two-qubit gates available on the quantum computer. By default,
        the gates are those in DEFAULT_TWO_QUBIT_GATES.
    reset_gates : Optional[Set[Type[OneQubitResetGate]]]
        Set of the reset gates available on the quantum computer. By default, the
        gates are those in DEFAULT_RESET_GATES.
    measurement_gates : Optional[Set[Type[OneQubitMeasurementGate]]]
        Set of the one-qubit measurement gates available on the quantum computer.
        By default, the gates are those in DEFAULT_MEASUREMENT_GATES.
    """

    def __init__(
        self,
        one_qubit_gates: set[type[OneQubitCliffordGate]] | None = None,
        two_qubit_gates: set[type[TwoOperandGate]] | None = None,
        reset_gates: set[type[OneQubitResetGate]] | None = None,
        measurement_gates: set[type[_MeasurementGate]] | None = None,
    ):
        one_qubit_gates_and_times = (
            dict.fromkeys(one_qubit_gates, 1.0)
            if one_qubit_gates is not None
            else None
        )
        two_qubit_gates_and_times = (
            dict.fromkeys(two_qubit_gates, 1.0)
            if two_qubit_gates is not None
            else None
        )
        reset_gates_and_times = (
            dict.fromkeys(reset_gates, 1.0) if reset_gates is not None else None
        )
        measurement_gates_and_times = (
            dict.fromkeys(measurement_gates, 1.0)
            if measurement_gates is not None
            else None
        )

        super().__init__(
            one_qubit_gates=one_qubit_gates_and_times,
            two_qubit_gates=two_qubit_gates_and_times,
            reset_gates=reset_gates_and_times,
            measurement_gates=measurement_gates_and_times,
        )


class ExhaustiveGateSet(NativeGateSet):
    """
    Class for capturing gateset of a quantum computer that can perform all gates
    natively.
    """

    def __init__(self):
        super().__init__(
            one_qubit_gates=ONE_QUBIT_GATES,
            two_qubit_gates=TWO_QUBIT_GATES,
            reset_gates=RESET_GATES,
            measurement_gates=MEASUREMENT_GATES
        )



# (c) Copyright Riverlane 2020-2025..
import importlib.metadata

from deltakit_explorer import analysis
from deltakit_explorer._api._client import Client
from deltakit_explorer._utils._logging import Logging

__version__ = importlib.metadata.version(__package__)

# Prevent import of `importlib` (and any other non-public objects) from this module.
del importlib

# List only public members in `__all__`.
__all__ = [s for s in dir() if not s.startswith("_")]



# (c) Copyright Riverlane 2020-2025.
"""
This module contains common implementation parts for planar codes.
Other planar code classes derive from PlanarCode.
"""
import itertools
import warnings
from abc import ABC, abstractmethod
from enum import Enum
from functools import cached_property
from pathlib import Path

import matplotlib.pyplot as plt
import numpy as np
from deltakit_circuit import PauliX, PauliZ, Qubit
from deltakit_circuit._basic_types import Coord2D, Coord2DDelta
from deltakit_circuit._qubit_identifiers import PauliGate

from deltakit_explorer.codes._css._css_code import CSSCode
from deltakit_explorer.codes._stabiliser import Stabiliser
from deltakit_explorer.enums._basic_enums import DrawingColours


class ScheduleType(Enum):
    """
    Enum for specifying the schedule type for measuring stabilisers for a planar
    code.
    There are five options available:

    - SIMULTANEOUS, meaning all the stabilisers will be measured at the same
      time step.
    - Z_FIRST, meaning all the Z stabilisers will be measured in the first
      time step and the X stabilisers in the second time step.
    - X_FIRST, meaning all the X stabilisers will be measured in the first
      time step and the Z stabilisers in the second time step.
    - Z_ONLY, meaning only the Z stabilisers will be measured.
    - X_ONLY, meaning only the X stabilisers will be measured.

    """

    SIMULTANEOUS = "simultaneous"
    Z_FIRST = "z_first"
    X_FIRST = "x_first"
    Z_ONLY = "z_only"
    X_ONLY = "x_only"


class PlanarCode(CSSCode, ABC):
    """
    Class representing an abstract planar CSS code.

    This class also contains methods that help set up circuits for memory
    experiments.

    Parameters
    ----------
    width: int
        Width of the code patch.
    height: int
        Height of the code patch.
    schedule_type : Optional[ScheduleType]
        The schedule type for measuring stabilisers. By default,
        ScheduleType.SIMULTANEOUS.
    use_ancilla_qubits: bool, optional
        Specifies whether the stabilisers should be defined with (True) or without
        (False) ancilla qubits. By default, True.
    shift : Coord2DDelta, optional
        A vector by which to shift all of the qubit coordinates, by adding it to each
        qubit coordinate.
        By default, the 0 vector: Coord2DDelta(0, 0).
    """

    def __init__(
        self,
        width: int,
        height: int,
        untransformed_x_schedule: tuple[Coord2DDelta, ...],
        untransformed_z_schedule: tuple[Coord2DDelta, ...],
        schedule_type: ScheduleType = ScheduleType.SIMULTANEOUS,
        use_ancilla_qubits: bool = True,
        shift: Coord2DDelta = Coord2DDelta(0, 0)
    ):
        self._shift = shift
        self.linear_tr = np.array([[1, 0], [0, 1]])
        if np.linalg.det(self.linear_tr) == 0.0:
            msg = "Determinant of linear transformation matrix cannot be 0"
            raise ValueError(msg)

        self._dimensions_at_least_2(width, height)

        self.width = width
        self.height = height

        self.schedule_type = schedule_type
        self._x_schedule = tuple(
            Coord2DDelta(*(self.linear_tr @ schedule_delta))
            for schedule_delta in untransformed_x_schedule
        )
        self._z_schedule = tuple(
            Coord2DDelta(*(self.linear_tr @ schedule_delta))
            for schedule_delta in untransformed_z_schedule
        )

        (
            self._untransformed_x_ancilla_qubits,
            self._untransformed_z_ancilla_qubits,
            self._untransformed_data_qubits,
        ) = self._calculate_untransformed_all_qubits()

        self._data_qubits = self._calculate_data_qubits()
        (
            self._x_ancilla_qubits,
            self._z_ancilla_qubits,
        ) = self._calculate_x_and_z_ancilla_qubits()

        stabilisers = self._calculate_stabilisers()
        (
            x_logicals,
            z_logicals,
        ) = self._calculate_logical_operators()

        self._perform_css_checks = getattr(self, "_perform_css_checks", True)

        super().__init__(
            stabilisers=stabilisers,
            x_logical_operators=x_logicals,
            z_logical_operators=z_logicals,
            use_ancilla_qubits=use_ancilla_qubits,
        )

    @staticmethod
    def _dimensions_at_least_2(width: int, height: int):
        """
        Verify both dimensions are greater than or equal to 2.

        Parameters
        ----------
        width: int
            Width of the code patch.
        height: int
            Height of the code patch.

        Raises
        ------
        ValueError
            Raises an error if either width or height is smaller than 2.
        """
        if width < 2 or height < 2:
            msg = "Width and height need to be greater than or equal to 2."
            raise ValueError(msg)

    @abstractmethod
    def _calculate_untransformed_logical_operators(
        self,
    ) -> tuple[tuple[set[PauliGate], ...], tuple[set[PauliGate], ...]]:
        """
        Calculate logical operators before the linear transformation and
        shift vector are applied.

        Returns
        -------
        Tuple[Tuple[Set[PauliGate], ...], Tuple[Set[PauliGate], ...]]
            Logical X operators, each consisting of X operators defining the
            logical and logical Z operators, each consisting of Z operators
            defining the logical.
        """

    def _calculate_logical_operators(
        self,
    ) -> tuple[tuple[set[PauliGate], ...], tuple[set[PauliGate], ...]]:
        """
        Calculate logical operators, applying a linear transformation
        and shift vector to each qubit coordinate.

        Returns
        -------
        Tuple[Tuple[Set[PauliGate], ...], Tuple[Set[PauliGate], ...]]
            Logical X operators, each consisting of X operators defining the
            logical and logical Z operators, each consisting of Z operators
            defining the logical.
        """
        (
            untransformed_x_logicals,
            untransformed_z_logicals,
        ) = self._calculate_untransformed_logical_operators()
        return tuple(
            set(
                {
                    PauliX(self._update_qubit_coordinates(pauli.qubit))
                    for pauli in logical
                }
            )
            for logical in untransformed_x_logicals
        ), tuple(
            set(
                {
                    PauliZ(self._update_qubit_coordinates(pauli.qubit))
                    for pauli in logical
                }
            )
            for logical in untransformed_z_logicals
        )

    @abstractmethod
    def _calculate_untransformed_all_qubits(
        self,
    ) -> tuple[set[Qubit], set[Qubit], set[Qubit]]:
        """
        Calculate X and Z-type stabiliser ancilla qubits and data qubits before
        the linear transformation and shift vector are applied.

        Returns
        -------
        Tuple[Set[Qubit], Set[Qubit], Set[Qubit]]
            Ancilla qubits used to measure X-type stabilisers, ancilla qubits
            used to measure Z-type stabilisers and data qubits.

        """

    def _calculate_x_and_z_ancilla_qubits(self) -> tuple[set[Qubit], set[Qubit]]:
        """
        Calculate X and Z-type stabiliser ancilla qubits, applying a linear transformation
        and shift vector to each qubit coordinate.

        Returns
        -------
        Tuple[Set[Qubit], Set[Qubit]]
            Ancilla qubits used to measure X-type stabilisers and ancilla
            qubits used to measure Z-type stabilisers.
        """
        return {
            self._update_qubit_coordinates(x_anc)
            for x_anc in self._untransformed_x_ancilla_qubits
        }, {
            self._update_qubit_coordinates(z_anc)
            for z_anc in self._untransformed_z_ancilla_qubits
        }

    def _calculate_data_qubits(self) -> set[Qubit]:
        """
        Calculate all data qubits and return a set of them, applying a
        linear transformation and shift vector to all the qubit coordinates.

        Returns
        -------
        Set[Qubit]
            Set of all data qubits.
        """
        return {
            self._update_qubit_coordinates(data_q)
            for data_q in self._untransformed_data_qubits
        }

    def _update_qubit_coordinates(self, qubit: Qubit) -> Qubit:
        """
        Given a qubit, a linear transformation and a shift vector, return a new qubit
        with coordinates acquired by adding the shift vector to the original qubit coordinates
        and then applying the linear transformation.

        Parameters
        ----------
        qubit : Qubit
            The qubit we wish to apply the transformations to.
        """
        qubit_coords_as_np = np.array(
            [qubit.unique_identifier.x, qubit.unique_identifier.y]
        )
        return Qubit(Coord2D(*(self.linear_tr @ qubit_coords_as_np)) + self._shift)

    def _calculate_ancilla_qubits(self) -> set[Qubit]:
        if self._use_ancilla_qubits:
            return self._x_ancilla_qubits.union(self._z_ancilla_qubits)
        return set()

    def _calculate_stabilisers(self) -> tuple[tuple[Stabiliser, ...], ...]:
        """
        Calculate stabilisers defining the code.

        Returns
        -------
        stabilisers: Tuple[Tuple[Stabiliser, ...], ...]
            Stabilisers defining the code. The ith element of stabilisers
            contains a tuple of stabilisers measured during the ith time step.
            Stabilisers measured during the same time step are also sorted from
            ones that are finished measuring first (meaning more Nones at the
            end of the tuple) to ones that are finished measuring last (meaning
            fewer Nones at the end of the tuple).
        """
        x_stabilisers = self._calculate_single_type_stabilisers(
            ancilla_qubits=self._x_ancilla_qubits,
            schedule=self._x_schedule,
            gate=PauliX,
        )
        z_stabilisers = self._calculate_single_type_stabilisers(
            ancilla_qubits=self._z_ancilla_qubits,
            schedule=self._z_schedule,
            gate=PauliZ,
        )

        stabilisers = []
        if self.schedule_type == ScheduleType.SIMULTANEOUS:
            stabilisers.append(self._sort_stabilisers(x_stabilisers + z_stabilisers))
        elif self.schedule_type == ScheduleType.Z_FIRST:
            stabilisers.append(self._sort_stabilisers(z_stabilisers))
            stabilisers.append(self._sort_stabilisers(x_stabilisers))
        elif self.schedule_type == ScheduleType.X_FIRST:
            stabilisers.append(self._sort_stabilisers(x_stabilisers))
            stabilisers.append(self._sort_stabilisers(z_stabilisers))
        elif self.schedule_type == ScheduleType.Z_ONLY:
            stabilisers.append(self._sort_stabilisers(z_stabilisers))
        elif self.schedule_type == ScheduleType.X_ONLY:
            stabilisers.append(self._sort_stabilisers(x_stabilisers))
        else:
            msg = f"Unknown schedule type {self.schedule_type} encountered."
            raise ValueError(msg)

        return tuple(stabilisers)

    def _calculate_single_type_stabilisers(
        self,
        ancilla_qubits: set[Qubit],
        schedule: tuple[Coord2DDelta, ...],
        gate: type[PauliGate],
    ) -> tuple[Stabiliser, ...]:
        """
        Calculate stabilisers, which are products of a single type of gate.

        Parameters
        ----------
        ancilla_qubits : Set[Qubit]
            Ancilla qubits to use for measuring stabilisers.
        schedule : Tuple[Coord2DDelta]
            Defines order in which controlled Paulis are applied between
            ancilla qubits and data qubits in the syndrome extraction circuit.
        gate : Type[PauliGate]
            Stabilisers are products of gates of type gate.

        Returns
        -------
        stabilisers: Tuple[Stabiliser, ...]
            Stabilisers, which are products of the input gate.
        """
        stabilisers = []
        for ancilla in ancilla_qubits:
            paulis = []
            for delta in schedule:
                qubit = Qubit(Coord2D(*ancilla.unique_identifier) + delta)
                if qubit in self._data_qubits:
                    paulis.append(gate(qubit))
                else:
                    paulis.append(None)

            stabilisers.append(Stabiliser(paulis=paulis, ancilla_qubit=ancilla))

        return tuple(stabilisers)

    @staticmethod
    def _sort_stabilisers(
        stabilisers: tuple[Stabiliser, ...],
    ) -> tuple[Stabiliser, ...]:
        """
        Sort a tuple of Stabilisers starting with stabilisers whose measurements
        can be finished earliest.

        Parameters
        ----------
        stabilisers : Tuple[Stabiliser, ...]
            Stabilisers to sort.

        Returns
        -------
        Tuple[Stabiliser, ...]
            Stabilisers sorted starting with stabilisers whose measurements
            can be finished earliest.
        """

        def _count_nones_at_the_end(stabiliser: Stabiliser) -> int:
            """Count the number of Nones at the end of stabiliser.paulis."""

            counter = 0
            for pauli in stabiliser.paulis[::-1]:
                if pauli is None:
                    counter += 1
                else:
                    break

            return counter

        def _contains_x(stabiliser: Stabiliser) -> int:
            """Return 1 if there are any PauliXs in the stabiliser and 0 otherwise."""
            for pauli in stabiliser.paulis:
                if isinstance(pauli, PauliX):
                    return 1
            return 0

        return tuple(
            sorted(
                sorted(stabilisers, key=_contains_x, reverse=True),
                key=_count_nones_at_the_end,
                reverse=True,
            )
        )

    def draw_patch(self, filename: str | None = None, unrotated_code: bool = False) -> None:
        """
        Draw a picture of the planar code, optionally saving it to a .png file.

        Parameters
        ----------
        filename: str, optional
            Path to the file where to save the pictorial representation of the
            planar code stored in this class.
        """
        all_qubit_x_coords = [qubit.unique_identifier.x for qubit in self.qubits]
        all_qubit_y_coords = [qubit.unique_identifier.y for qubit in self.qubits]
        diff_from_max_coord_to_margin_no_ancilla = (
            2 if not unrotated_code or not (self.linear_tr == np.eye(2)).all() else 1
        )
        if self._use_ancilla_qubits:
            min_x, max_x = min(all_qubit_x_coords) - 1, max(all_qubit_x_coords) + 1
            min_y, max_y = min(all_qubit_y_coords) - 1, max(all_qubit_y_coords) + 1
        else:
            min_x, max_x = (
                min(all_qubit_x_coords) - diff_from_max_coord_to_margin_no_ancilla,
                max(all_qubit_x_coords) + diff_from_max_coord_to_margin_no_ancilla,
            )
            min_y, max_y = (
                min(all_qubit_y_coords) - diff_from_max_coord_to_margin_no_ancilla,
                max(all_qubit_y_coords) + diff_from_max_coord_to_margin_no_ancilla,
            )
        x_lim = (min_x, max_x)
        y_lim = (min_y, max_y)
        fig, ax = plt.subplots(nrows=1, ncols=1)
        ax.set_xlim(x_lim)
        ax.set_ylim(y_lim)

        stabilisers = tuple(itertools.chain.from_iterable(self._stabilisers))
        stabilisers = self._sort_stabilisers(stabilisers)

        # Draw stabiliser plaquettes
        for stabiliser in stabilisers:
            data_qubit_x_coords = [
                pauli.qubit.unique_identifier[0]
                for pauli in stabiliser.paulis
                if pauli is not None
            ]
            data_qubit_y_coords = [
                pauli.qubit.unique_identifier[1]
                for pauli in stabiliser.paulis
                if pauli is not None
            ]

            paulis = [pauli for pauli in stabiliser.paulis if pauli is not None]

            if len(paulis) == 2:
                ancilla_coord = stabiliser.ancilla_qubit.unique_identifier
                data_qubit_x_coords.append(ancilla_coord[0])
                data_qubit_y_coords.append(ancilla_coord[1])
            elif len(paulis) == 4:
                data_qubit_x_coords[2], data_qubit_x_coords[3] = (
                    data_qubit_x_coords[3],
                    data_qubit_x_coords[2],
                )
                data_qubit_y_coords[2], data_qubit_y_coords[3] = (
                    data_qubit_y_coords[3],
                    data_qubit_y_coords[2],
                )

            if isinstance(paulis[0], PauliX):
                ax.fill(
                    data_qubit_x_coords,
                    data_qubit_y_coords,
                    color=DrawingColours.X_COLOUR.value,
                    alpha=1,
                )
            else:
                ax.fill(
                    data_qubit_x_coords,
                    data_qubit_y_coords,
                    color=DrawingColours.Z_COLOUR.value,
                    alpha=1,
                )

        # Draw data qubits
        for qubit in self._data_qubits:
            cc = plt.Circle(
                qubit.unique_identifier,
                0.2,
                color=DrawingColours.DATA_QUBIT_COLOUR.value,
                alpha=1,
            )
            ax.set_aspect(1)
            ax.add_artist(cc)

        if self._use_ancilla_qubits:
            # Draw X stabiliser ancilla qubits
            for qubit in self._x_ancilla_qubits:
                cc = plt.Circle(
                    qubit.unique_identifier,
                    0.2,
                    color=DrawingColours.ANCILLA_QUBIT_COLOUR.value,
                    alpha=1,
                )
                ax.set_aspect(1)
                ax.add_artist(cc)

            # Draw Z stabiliser ancilla qubits
            for qubit in self._z_ancilla_qubits:
                cc = plt.Circle(
                    qubit.unique_identifier,
                    0.2,
                    color=DrawingColours.ANCILLA_QUBIT_COLOUR.value,
                    alpha=1,
                )
                ax.set_aspect(1)
                ax.add_artist(cc)

        # Save the file
        if filename:
            output_directory = Path(filename)
            if not output_directory.exists():
                output_directory.parent.mkdir(parents=True, exist_ok=True)
            fig.savefig(filename)
            plt.close(fig)

    @cached_property
    def x0(self) -> int:
        """
        Across the data qubits of the patch, get the smallest x-coordinate value.

        Returns
        -------
        int
            The smallest x-coordinate value across all the data qubits in the patch.
        """
        return min(
            self.data_qubits, key=lambda qubit: qubit.unique_identifier.x
        ).unique_identifier.x

    @cached_property
    def x1(self) -> int:
        """
        Across the data qubits of the patch, get the largest x-coordinate value.

        Returns
        -------
        int
            The largest x-coordinate value across all the data qubits in the patch.
        """
        return max(
            self.data_qubits, key=lambda qubit: qubit.unique_identifier.x
        ).unique_identifier.x

    @cached_property
    def y0(self) -> int:
        """
        Across the data qubits of the patch, get the smallest y-coordinate value.

        Returns
        -------
        int
            The smallest y-coordinate value across all the data qubits in the patch.
        """
        return min(
            self.data_qubits, key=lambda qubit: qubit.unique_identifier.y
        ).unique_identifier.y

    @cached_property
    def y1(self) -> int:
        """
        Across the data qubits of the patch, get the largest y-coordinate value.

        Returns
        -------
        int
            The largest y-coordinate value across all the data qubits in the patch.
        """
        return max(
            self.data_qubits, key=lambda qubit: qubit.unique_identifier.y
        ).unique_identifier.y

    def get_shifted_logicals(
        self, shift_z: bool, delta: Coord2DDelta
    ) -> tuple[frozenset[PauliGate], ...]:
        """
        Given a boolean specifying whether to modify X or Z logicals, and a Coord2DDelta,
        modify the relevant logical(s) by adding the Coord2DDelta to the coordinates of
        each qubit making up the logical.
        A check is performed to ensure the shifted logical still lies on the data qubits
        of the patch. If this is not the case, a ValueError is thrown.

        Parameters
        ----------
        shift_z : bool
            Boolean specifying whether to shift the Z-logicals (True) or X-logicals
            (False).
        delta : Coord2DDelta
            The delta to add to each qubit coordinate.

        Returns
        -------
        tuple[frozenset[PauliGate], ...]
            A tuple of frozensets of PauliX or PauliZ operators, depending on the
            value of shift_z being False or True respectively, which will be the
            logicals of the code modified by adding the delta vector to each qubit
            coordinate.
        """
        if shift_z:
            logicals = self.z_logical_operators
            pauli_op = PauliZ
        else:
            logicals = self.x_logical_operators
            pauli_op = PauliX
        new_logicals = []
        for old_logical in logicals:
            new_logical = frozenset(
                pauli_op(Qubit(pauli.qubit.unique_identifier + delta))
                for pauli in old_logical
            )
            # Verify that the new logical lies within the patch, if we are not the
            # UnrotatedToricCode class. Do it with __name__ to avoid circular import
            if type(self).__name__ != "UnrotatedToricCode":
                for pauli in new_logical:
                    if pauli.qubit not in self.data_qubits:
                        msg = f"Pauli operator on qubit {pauli.qubit} is not contained in the patch data qubits after shifting by {tuple(delta)}"
                        raise ValueError(msg)
            else:
                warnings.warn(
                    "Logical shifting validation is disabled for the UnrotatedToricCode; new logicals may be invalid"
                )
            new_logicals.append(new_logical)
        return tuple(new_logicals)



# (c) Copyright Riverlane 2020-2025.
"""
This module provides classes to define quantum memory experiments for
planar unrotated and rotated codes. It is not currently public; this
__init__.py file is a historical artefact and can be removed, adjusting
imports within other `deltakit_explorer` modules accordingly.
"""

from deltakit_explorer.codes._planar_code._planar_code import ScheduleType
from deltakit_explorer.codes._planar_code._rotated_planar_code import RotatedPlanarCode
from deltakit_explorer.codes._planar_code._unrotated_planar_code import (
    UnrotatedPlanarCode,
)
from deltakit_explorer.codes._planar_code._unrotated_toric_code import (
    UnrotatedToricCode,
)

__all__ = [s for s in dir() if not s.startswith("_")]



# (c) Copyright Riverlane 2020-2025.
"""
This module includes implementation of a SI1000 noise model.
Toy noise is a default choice of noise parameters to explore codes
scalability and to compare against other works. This a de-facto
experimental standard.
"""

from dataclasses import dataclass

from deltakit_circuit import after_reset_flip_probability, measurement_noise_profile
from deltakit_circuit.gates import TWO_QUBIT_GATES, OneQubitCliffordGate
from deltakit_circuit.noise_channels import Depolarise1, Depolarise2

from deltakit_explorer.qpu._noise._leakage_noise_profiles import (
    idle_qubit_relaxation_noise_profile,
    one_qubit_clifford_gate_relaxation_noise_profile,
    qubit_reset_leakage_noise_profile,
    resonator_idle_qubit_relaxation_noise_profile,
    two_qubit_gate_leakage_noise_profile,
    two_qubit_gate_relaxation_noise_profile,
)
from deltakit_explorer.qpu._noise._noise_parameters import NoiseParameters


@dataclass(kw_only=True)
class SI1000Noise(NoiseParameters):
    """
    Superconducting inspired noise model from: https://arxiv.org/abs/2108.10457.
    This noise model assumes that after every measurement there is a reset.
    Includes an optional parameter pL for leakage implemented as described
    in the LCD paper: https://arxiv.org/abs/2411.10343.

    Parameters
    ----------
    p: float
        Physical error rate.
    pL: float
        Informs leakage related probabilities.
    """

    name = "SI1000_noise"

    p: float = 0
    pL: float = 0

    def __post_init__(self):
        depolarise1_generator = Depolarise1.generator_from_prob(self.p / 10)
        depolarise2_generator = Depolarise2.generator_from_prob(self.p)

        self.gate_noise.append(
            lambda noise_context: depolarise1_generator(
                noise_context.gate_layer_qubits(
                    (*tuple(TWO_QUBIT_GATES), OneQubitCliffordGate), gate_qubit_count=1
                )
            )
        )

        self.gate_noise.append(
            lambda noise_context: depolarise2_generator(
                noise_context.gate_layer_qubits(
                    tuple(TWO_QUBIT_GATES), gate_qubit_count=2
                )
            )
        )

        self.idle_noise = lambda qubit, _t=0.0: Depolarise1(
            qubit=qubit, probability=self.p / 10
        )

        self.reset_noise = after_reset_flip_probability(2 * self.p)

        resonator_leakage = 4 * self.p if self.pL > 0 else 0
        self.gate_noise.append(
            resonator_idle_qubit_relaxation_noise_profile(2 * self.p, resonator_leakage)
        )

        if self.pL > 0:
            self.gate_noise.append(two_qubit_gate_leakage_noise_profile(self.pL))
            self.reset_noise.append(qubit_reset_leakage_noise_profile(self.pL))
            self.gate_noise.append(two_qubit_gate_relaxation_noise_profile(self.pL))
            self.gate_noise.append(
                one_qubit_clifford_gate_relaxation_noise_profile(self.p / 5)
            )
            self.gate_noise.append(idle_qubit_relaxation_noise_profile(self.p / 5))

        self.measurement_flip = measurement_noise_profile(5 * self.p)

    def __str__(self) -> str:
        if self.pL:
            return f"{self.name}_{self.p:.0e}_{self.pL:.0e}"
        return f"{self.name}_{self.p:.0e}"



# (c) Copyright Riverlane 2020-2025.
"""
Collection of dictionaries for different types of schedules for Rotated Planar
codes.
"""

from dataclasses import dataclass
from typing import ClassVar

from deltakit_circuit._basic_types import Coord2DDelta


@dataclass
class RotatedPlanarCodeSchedules:
    """
    Dataclass to capture default syndrome extraction schedules for the rotated
    planar code.
    """

    # In default_N_Z_dict,
    #     "N" means the N-shaped schedule, i.e. for that plaquette we use the
    #     following schedule:
    #             ---------------
    #             |             |
    #             |   ||    |   |
    #             |   | |   ˅   |
    #             |   ˅  |  |   |
    #             |   |   | |   |
    #             |             |
    #             ---------------
    #     and "Z" means the Z-shaped schedule, i.e. for that plaquette we use the
    #     following schedule:
    #             --------------
    #             |            |
    #             |    --<--   |
    #             |    |       |
    #             |     |      |
    #             |      |     |
    #             |  --<--     |
    #             |            |
    #             --------------
    default_N_Z_dict: ClassVar[
        dict[str, tuple[Coord2DDelta, Coord2DDelta, Coord2DDelta, Coord2DDelta]]
    ] = {
        "N": (
            Coord2DDelta(1, 1),
            Coord2DDelta(1, -1),
            Coord2DDelta(-1, 1),
            Coord2DDelta(-1, -1),
        ),
        "Z": (
            Coord2DDelta(1, 1),
            Coord2DDelta(-1, 1),
            Coord2DDelta(1, -1),
            Coord2DDelta(-1, -1),
        ),
    }

    # Horizontally reflected version of default_N_Z_dict, i.e. with
    #         ---------------
    #         |             |
    #         |   |    ||   |
    #         |   |   | ˅   |
    #         |   ˅  |  |   |
    #         |   | |   |   |
    #         |             |
    #         ---------------
    # and
    #         --------------
    #         |            |
    #         |    -->--   |
    #         |       |    |
    #         |     |      |
    #         |   |        |
    #         |  -->--     |
    #         |            |
    #         --------------
    horizontally_reflected_N_Z_dict: ClassVar[
        dict[str, tuple[Coord2DDelta, Coord2DDelta, Coord2DDelta, Coord2DDelta]]
    ] = {
        "N": (
            Coord2DDelta(-1, 1),
            Coord2DDelta(-1, -1),
            Coord2DDelta(1, 1),
            Coord2DDelta(1, -1),
        ),
        "Z": (
            Coord2DDelta(-1, 1),
            Coord2DDelta(1, 1),
            Coord2DDelta(-1, -1),
            Coord2DDelta(1, -1),
        ),
    }

    # Vertically reflected version of default_N_Z_dict, i.e. with
    #         ---------------
    #         |             |
    #         |   |   | |   |
    #         |   |  |  |   |
    #         |   ^ |   ^   |
    #         |   ||    |   |
    #         |             |
    #         ---------------
    # and
    #         --------------
    #         |            |
    #         |    --<--   |
    #         |       |    |
    #         |     |      |
    #         |   |        |
    #         |  --<--     |
    #         |            |
    #         --------------
    vertically_reflected_N_Z_dict: ClassVar[
        dict[str, tuple[Coord2DDelta, Coord2DDelta, Coord2DDelta, Coord2DDelta]]
    ] = {
        n_or_z: schedule[::-1]
        for n_or_z, schedule in horizontally_reflected_N_Z_dict.items()
    }

    # Doubly reflected version of default_N_Z_dict, i.e. with
    #         ---------------
    #         |             |
    #         |   ||    |   |
    #         |   | |   ^   |
    #         |   ^  |  |   |
    #         |   |   | |   |
    #         |             |
    #         ---------------
    # and
    #         --------------
    #         |            |
    #         |    -->--   |
    #         |    |       |
    #         |     |      |
    #         |      |     |
    #         |  -->--     |
    #         |            |
    #         --------------
    doubly_reflected_N_Z_dict: ClassVar[
        dict[str, tuple[Coord2DDelta, Coord2DDelta, Coord2DDelta, Coord2DDelta]]
    ] = {n_or_z: schedule[::-1] for n_or_z, schedule in default_N_Z_dict.items()}



from collections.abc import Callable
from dataclasses import dataclass

from deltakit_circuit import measurement_noise_profile
from deltakit_circuit.gates import (
    I,
    OneQubitCliffordGate,
    OneQubitMeasurementGate,
    OneQubitResetGate,
)
from deltakit_circuit.noise_channels import Depolarise1, Depolarise2

from deltakit_explorer.qpu._noise._noise_parameters import (
    NoiseParameters,
    _idle_noise_from_t1_t2,
)


@dataclass(kw_only=True)
class PhysicalNoise(NoiseParameters):
    r"""Gets gate noise parameters given physical gate data

    Parameters
    ----------
    t1: float
        :math:`T_1` time (relaxation from :math:`\ket{1}` to :math:`\ket{0}`), seconds.
    t2: float
        :math:`T_2` time (dephasing), seconds.
    p_1_qubit_gate_error: float
        Probability of a flip while doing a 1-qubit gate.
    p_2_qubit_gate_error: float
        Probability of a flip while doing a 2-qubit gate.
    p_reset_error: float
        Probability of a flip while doing a reset.
    p_meas_qubit_error: float
        Probability of incorrect measurement.
    p_readout_flip: float
        Probability of a flip while measuring a qubit.

    Returns
    -------
    noise_model: NoiseParameters
        A `NoiseParameters` object representing the noise model of a QPU.

    Examples
    --------
    >>> from deltakit.circuit import Qubit
    >>> from deltakit.explorer import qpu
    >>> noise_model = qpu.PhysicalNoise(
    ...     t1=20e-6,
    ...     t2=30e-6,
    ...     p_readout_flip=0.01,
    ...     p_1_qubit_gate_error=0.001,
    ...     p_2_qubit_gate_error=0.01,
    ...     p_reset_error=0.01,
    ...     p_meas_qubit_error=0.01,
    ... )

    """

    name = "PhysicalNoise"

    t1: float = 0
    t2: float = 0
    p_1_qubit_gate_error: float = 0
    p_2_qubit_gate_error: float = 0
    p_reset_error: float = 0
    p_meas_qubit_error: float = 0
    p_readout_flip: float = 0

    def __post_init__(self) -> None:
        idle_noise = _idle_noise_from_t1_t2(self.t1, self.t2)

        def _gate_noise(noise_context):
            noise_ops = []
            for gate in noise_context.gate_layer.gates:
                if isinstance(gate, I):
                    continue
                if len(gate.qubits) == 2:
                    noise_ops.append(
                        Depolarise2(*gate.qubits, self.p_2_qubit_gate_error)  # type: ignore[call-arg]
                    )
                elif isinstance(gate, OneQubitCliffordGate):
                    noise_ops.append(
                        Depolarise1(*gate.qubits, self.p_1_qubit_gate_error)
                    )
            return noise_ops

        gate_noise = [_gate_noise]

        reset_noise: list[Callable] = [
            lambda noise_context: Depolarise1.generator_from_prob(self.p_reset_error)(
                noise_context.gate_layer_qubits(OneQubitResetGate)
            )
        ]
        measurement_noise: list[Callable] = [
            lambda noise_context: Depolarise1.generator_from_prob(
                self.p_meas_qubit_error
            )(noise_context.gate_layer_qubits(OneQubitMeasurementGate))
        ]
        measurement_flip_noise = measurement_noise_profile(self.p_readout_flip)

        self.gate_noise = gate_noise
        self.idle_noise = idle_noise
        self.reset_noise = reset_noise
        self.measurement_noise_after = measurement_noise
        self.measurement_flip = measurement_flip_noise



# (c) Copyright Riverlane 2020-2025.
"""
This module introduces named gate sets for convenient type checks
and default initialisation.
"""


from deltakit_circuit.gates import CX, MZ, RZ, H, S, X, Y, Z

DEFAULT_ONE_QUBIT_GATES = {X, Y, Z, H, S}

DEFAULT_TWO_QUBIT_GATES = {CX}

DEFAULT_MEASUREMENT_GATES = {MZ}

DEFAULT_RESET_GATES = {RZ}



# (c) Copyright Riverlane 2020-2025.
"""This module accumulates classes for data transformation.
Contantins converters between most popular formats: 01, b8, and csv"""
from __future__ import annotations

from codecs import StreamReader, StreamWriter
from collections.abc import Iterable
from io import BufferedIOBase, BytesIO, RawIOBase, StringIO, TextIOWrapper
from pathlib import Path

import numpy.typing as npt

from deltakit_explorer.enums._api_enums import DataFormat


def read_01(stream: StreamReader | TextIOWrapper) -> Iterable[list[int]]:
    """Read 01 data from steam, StringIO or file descriptor."""
    for line in stream:
        yield [int(char) for char in line.strip()]

def read_csv(stream: StreamReader | TextIOWrapper) -> Iterable[list[int]]:
    """Read CSV data from steam, StringIO or file descriptor."""
    for line in stream:
        yield [int(char) for char in line.strip().split(",")]

def read_b8(
    stream: BufferedIOBase | RawIOBase,
    width: int
) -> Iterable[list[int]]:
    """Read bit-packed data from steam, BytesIO or file descriptor."""
    bytes_per_shot = (width + 7) // 8
    while (shot := stream.read(bytes_per_shot)):
        result = []
        # iterate through bits
        for k in range(width):
            which_byte = shot[k // 8]
            which_bit = (which_byte >> (k % 8)) % 2
            result.append(which_bit)
        yield result

def write_01(
    data: Iterable[list[int]] | npt.NDArray,
    stream: StreamWriter | TextIOWrapper | StringIO,
):
    """Write data into a text stream in 01 format."""
    for row in data:
        # convert to string, and then join with empty
        stream.write("".join(map(str, row)) + "\n")

def write_csv(
    data: Iterable[list[int]] | npt.NDArray,
    stream: StreamWriter | TextIOWrapper | StringIO,
):
    """Write data into a text stream in CSV format."""
    for row in data:
        # convert to string, and then join with empty
        stream.write(",".join(map(str, row)) + "\n")

def write_b8(
    data: Iterable[list[int]] | npt.NDArray,
    stream: BufferedIOBase | RawIOBase | BytesIO,
):
    """Write data into a bytes stream in B8 format."""
    for row in data:
        bytes_per_shot = (len(row) + 7) // 8
        longint = 0
        for bit in reversed(row):
            int_bit = int(bit)
            longint <<= 1
            longint += int_bit
        stream.write(longint.to_bytes(bytes_per_shot, "little"))

def write_binary_data_to_file(
    data: Iterable[list[int]] | npt.NDArray,
    destination_format: DataFormat,
    result_file: str | Path,
):
    """Writes a stream of bit lines (each List[int] is a line of {0, 1})
    into a file of desired format.

    Args:
        data (Union[Iterable[List[int]], npt.NDArray]):
            iterable of bit strings, or numpy array.
        destination_format (DataFormat):
            one of popular format, 01, b8, or csv.
        result_file (str | Path):
            str or Path object. Path to a file.
    """
    if destination_format == DataFormat.F01:
        with Path.open(Path(result_file), "w", encoding="utf-8") as string_stream:
            write_01(data, string_stream)

    elif destination_format == DataFormat.CSV:
        with Path.open(Path(result_file), "w", encoding="utf-8") as string_stream:
            write_csv(data, string_stream)

    elif destination_format == DataFormat.B8:
        with Path.open(Path(result_file), "wb") as byte_stream:
            write_b8(data, byte_stream)
    else:
        msg = f"Not supported format: {destination_format}"
        raise NotImplementedError(msg)
